// convf.C  Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 09:43:40 tac>
//
// dansのデータファイルを入力し、数字データの各種変換を行う。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W convf convf.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "convf  Version 7.41          Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:convf [\"-<変換式>\" …] [<オプション> …] [<データファイル> …]",
  "機能:指定された変換式に従って数字データの変換をするでござる。",
  "変換式:i[<変換後データ出力開始項目番号>][-[<終了項目番号>]]=<式>",
  "識別情報関数:item(),title()",
  "数学関数:abs(),cos(),exp(),int(),ln(),log(),rand(),sgn(),sin(),sqrt(),tan()",
  "統計関数:max(),mean(),min(),n(),sd(),sum(),tmax(),tmin()",
  "<オプション>",
  "  -c=[<開始ケース番号>][-[<終了ケース番号>]]:変換ケース番号の指定。",
  "  -cd[={-|+}]:ケースデータの出力指定でござる。(-cd:-cd=+,オプション省略時:-cd=-)",
  "  -f=<ファイル名>:変換式やオプションを指定したファイルから読み込み申す。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示し申す。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定でござる。(-id:-id=+)",
  "  -prec[=<有効数字>]:変換後データの有効数字の指定でござる。(0≦有効数字≦32)",
  "  -r<層別項目番号>=[<実数下限>][-[<実数上限>]]:実数層別項目指定",
  "  -s<層別項目番号>=<文字列>[,<文字列>,…]:文字層別項目指定",
  "    指定した実数範囲内または文字列と一致するケースだけ変換するでござる。"
};

constexpr int PRECdef = 16;     // 省略時変換後データの有効数字
constexpr int PRECmax = 32;     // 変換後データの有効数字最大値

enum idtype                     // 因子(数字・項目・関数)と演算子の種類
{
  idnop,                        // ヌル演算子(level=0)
  idplusop,                     // 単項演算子+(level=1)
  idminusop,                    // 単項演算子-(level=1)
  idadd,                        // 二項演算子+(level=1)
  idsub,                        // 二項演算子-(level=1)
  idmulti,                      // 二項演算子*(level=2)
  iddivd,                       // 二項演算子/(level=2)
  idmod,                        // 二項演算子%(level=2)
  idpower,                      // 二項演算子^(level=3)
  idfn,                         // 関数()(level=4)
  idfnabs,                      // 関数abs()(level=4)
  idfncos,                      // 関数cos()(level=4)
  idfnexp,                      // 関数exp()(level=4)
  idfnint,                      // 関数int()(level=4)
  idfnitem,                     // 項目名変換関数item()(level=1)
  idfnln,                       // 関数ln()(level=4)
  idfnlog,                      // 関数log()(level=4)
  idfnsgn,                      // 関数sgn()(level=4)
  idfnsin,                      // 関数sin()(level=4)
  idfnsqrt,                     // 関数sqrt()(level=4)
  idfntan,                      // 関数tan()(level=4)
  idfnmax,                      // 関数max()(level=4)
  idfnmean,                     // 関数mean()(level=4)
  idfnmin,                      // 関数min()(level=4)
  idfnn,                        // 関数n()(level=4)
  idfnrand,                     // 関数rand()(level=4)
  idfnsd,                       // 関数sd()(level=4)
  idfnsum,                      // 関数sum()(level=4)
  idfntitle,                    // データ名変換関数title()(level=1)
  idfntmax,                     // 関数tmax()(level=4)
  idfntmin,                     // 関数tmin()(level=4)
  idnum,                        // 数字と項目num()(level=4)
  iditem,                       // 数字と項目item()(level=4)
  idblank                       // ブランクデータ
};

enum cargtype                   // 引数の種類
{
  cargblank,                    // 引数なし
  itemsub,                      // 項目
  itemapp,                      // 項目名追加
  sdata,                        // 文字列データ
  cargetc                       // その他
};

class carg                      // 引数クラス
{
private:
  cargtype vtype;               // 引数の種類
  pairint irange;               // 変換項目の開始・終了番号
  string vstr1;                 // 第1文字列(変換対象文字列等)
  string vstr2;                 // 第2文字列(出力指定・変換後文字列等)
  bool vglobal;                 // グローバル置換フラッグ
  char vctype;                  // 変換指定型('d','f','g','s','x'等)
  pairdbl frange;               // 実数の上・下限(範囲指定等)
public:
  carg() : vtype(cargblank), irange(0, 0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // デフォールトコンストラクタ
  {}
  explicit carg(cargtype type0, int is0 = 0, int ie0 = 0) : vtype(type0), irange(is0, ie0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // コンストラクタ
  {}
  carg(cargtype type0, int is0, int ie0, const string& str10, const string& str20, bool global0 = false, char ctype0 = ' ', double low0 = BLANKf, double up0 = BLANKf) : vtype(type0), irange(is0, ie0), vstr1(str10), vstr2(str20), vglobal(global0), vctype(ctype0), frange(low0, up0)
  // コンストラクタ
  {}
  carg(const carg& carg0) : vtype(carg0.vtype), irange(carg0.irange), vstr1(carg0.vstr1), vstr2(carg0.vstr2), vglobal(carg0.vglobal), vctype(carg0.vctype), frange(carg0.frange)
  // コピーコンストラクタ
  {}
  virtual ~carg()               // デストラクタ
  {}
  const carg& operator=(const carg& carg0); // コピー代入演算子
  cargtype type() const                     // 引数の種類を返す
  { return vtype; }
  cargtype type(cargtype type0) // 引数の種類を設定する
  { return (vtype = type0); }
  int is() const                // 変換項目の開始番号を返す
  { return irange.first; }
  int is(int is0)               // 変換項目の開始番号を設定する
  { return (irange.first = is0); }
  int ie() const                // 変換項目の終了番号を返す
  { return irange.second; }
  int ie(int ie0)               // 変換項目の終了番号を設定する
  { return (irange.second = ie0); }
  const string& str1() const    // 文字列1を返す
  { return vstr1; }
  const string& str1(const string& str10) // 文字列1を設定する
  { return (vstr1 = str10); }
  const string& str2() const    // 文字列2を返す
  { return vstr2; }
  const string& str2(const string& str20) // 文字列2を設定する
  { return (vstr2 = str20); }
  bool global() const           // グローバル置換フラッグを返す
  { return vglobal; }
  bool global(bool global0)     // グローバル置換フラッグを設定する
  { return (vglobal = global0); }
  char ctype() const            // 変換指定型を返す
  { return vctype; }
  char ctype(char ctype0)       // 変換指定型を設定する
  { return (vctype = ctype0); }
  double low() const            // 実数の下限を返す
  { return frange.first; }
  double low(double low0)       // 実数の下限を設定する
  { return (frange.first = low0); }
  double up() const             // 実数の上限を返す
  { return frange.second; }
  double up(double up0)         // 実数の上限を設定する
  { return (frange.second = up0); }
};

const carg& carg::operator=(const carg& carg0) // コピー代入演算子
{
  if (&carg0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = carg0.vtype;  irange = carg0.irange;
  vstr1 = carg0.vstr1;  vstr2 = carg0.vstr2;  vglobal = carg0.vglobal;
  vctype = carg0.vctype;  frange = carg0.frange;
  return *this;
}

class term                      // 変換式の項クラス
{
private:
  idtype vtype;                 // 項の識別子型
  pairint vino;                 // 項目の開始・終了番号
  double vf;                    // 実数データ
  term* vfact;                  // 因子へのポインター
  term *vpre, *vnext;           // 前後の項へのポインター
  list<carg> vcarglst;          // 引数リスト
  int get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr) const;
  // 置換対象文字列と置換後文字列とグローバル指定を取得する
  cargtype get_target(int& is, int& ie, string& targstr, const string& sstr);
  // 変換対象項目または変換対象文字列を取得する
public:
  term() : vtype(idblank), vino(0, 0), vf(BLANKf), vfact(0), vpre(0), vnext(0), vcarglst()
  // デフォールトコンストラクタ
  {}
  term(idtype type0, int is0 = 0, int ie0 = 0, double f0 = BLANKf) : vtype(type0), vino(is0, ie0), vf(f0), vfact(0), vpre(0), vnext(0), vcarglst()
  // コンストラクタ
  {}
  term(const term& term0);      // コピーコンストラクタ
  virtual ~term();              // デストラクタ
  const term& operator=(const term& term0); // コピー代入演算子
  idtype type() const                       // 項の識別子型を返す
  { return vtype; }
  idtype type(idtype type0)     // 項の識別子型を設定する
  { return (vtype = type0); }
  int is() const                // 項目の開始番号を返す
  { return vino.first; }
  int is(int is0)               // 項目の開始番号を設定する
  { return (vino.first = is0); }
  int ie() const                // 項目の終了番号を返す
  { return vino.second; }
  int ie(int ie0)               // 項目の終了番号を設定する
  { return (vino.second = ie0); }
  double f() const              // 実数データを返す
  { return vf; }
  double f(double f0)           // 実数データを設定する
  { return (vf = f0); }
  term* fact() const            // 因子へのポインターを返す
  { return vfact; }
  term* fact(term* fact0)       // 因子へのポインターを設定する
  { return (vfact = fact0); }
  term* pre() const             // 前の項へのポインターを返す
  { return vpre; }
  term* pre(term* pre0)         // 前の項へのポインターを設定する
  { return (vpre = pre0); }
  term* next() const            // 後の項へのポインターを返す
  { return vnext; }
  term* next(term* next0)       // 後の項へのポインターを設定する
  { return (vnext = next0); }
  list<carg>& carglst()         // 引数リストを返す
  { return vcarglst; }
  int get_sub(const string& pname, const string& cargstr);
  // item()関数の引数を取得する(置換文字列指定)
};

term::term(const term& term0) : vtype(term0.vtype), vino(term0.vino), vf(term0.vf), vfact(0), vpre(0), vnext(0), vcarglst(term0.vcarglst)
// コピーコンストラクタ(オブジェクトを再帰的にコピーする)
{
  if (term0.vfact != 0) vfact = new term(*(term0.vfact));
  if (term0.vpre != 0) vpre = new term(*(term0.vpre));
  if (term0.vnext != 0) vnext = new term(*(term0.vnext));
}

term::~term()                   // デストラクタ
{
  delete vfact;
  delete vpre;
  delete vnext;                 // 因子と前後の項を再帰的に削除する
}

const term& term::operator=(const term& term0) // コピー代入演算子
{
  if (&term0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = term0.vtype;  vino = term0.vino;  vf = term0.vf;
  vfact = 0;  vpre = 0;  vnext = 0;
  if (term0.vfact != 0) vfact = new term(*(term0.vfact));
  if (term0.vpre != 0) vpre = new term(*(term0.vpre));
  if (term0.vnext != 0) vnext = new term(*(term0.vnext));
  return *this;
}

int term::get_sub(const string& pname, const string& cargstr)
// title()とitem()関数の引数を取得する(置換文字列指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "item()の変換内容が間違ってござる→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  string targstr, pattern, repstr;
  bool global = false;
  if (get_pattern(targstr, pattern, repstr, global, *itr) < 0) return disperr(pname, "最初の変換指定が間違ってござる→", cargstr);
  int is = 0, ie = 0;
  cargtype atype = sdata;
  if (vtype != idfntitle)       // title()以外
  {
    string tstr;
    atype = get_target(is, ie, tstr, targstr);
    if (atype != itemsub && atype != itemapp) return disperr(pname, "item()の変換対象指定は項目だけでござる→", cargstr);
  }
  carg tag(atype, is, ie, pattern, repstr, global);
  vcarglst.push_back(tag);
  ++itr;
  int err = 0, n = 1;
  while (!err && itr != cargstrlst.end())
  {
    if (get_pattern(targstr, pattern, repstr, global, *itr) < 0)
    { err = disperr(pname, "2番目以後の変換指定が間違ってごさる→", cargstr); }
    else
    {
      tag.str1(pattern);  tag.str2(repstr);  tag.global(global);
      vcarglst.push_back(tag);
      ++itr;
    }
  }
  return err ? err : n;         // エラーでなければ置換数を返す
}

int term::get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr) const
// 置換対象文字列と置換後文字列とグローバル指定を取得する
{
  vecstr cargstrvec;
  if (strsplit(cargstrvec, cargstr, CHRstr) != 4) return -1;
  targstr = cargstrvec[0];  pattern = cargstrvec[1];  repstr = cargstrvec[2];
  sstrtostr(pattern);  sstrtostr(repstr); // 特殊表現文字列を変換し""を削除する
  global = (cargstrvec[3] == "g") ? true : false;
  return 0;
}

cargtype term::get_target(int& is, int& ie, string& targstr, const string& sstr)
// 変換対象項目または変換対象文字列を取得する
{
  targstr = sstr;
  strtrm(targstr);
  if (targstr.size() < 1) return cargblank;
  is = 0;  ie = 0;
  if (targstr[0] == 'i')        // 項目指定の場合
  {
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(1)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemsub;
  }
  if (targstr.size() > 1 && targstr[0] == '+' && targstr[1] == 'i')
  {                             // 項目名追加の場合
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(2)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemapp;
  }
  sstrtostr(targstr);           // 特殊表現文字を変換し、""を削除する
  return sdata;                 // 文字列指定の場合
}

class form                      // 変換式クラス
{
private:
  pairint vono;                 // 出力項目の開始・終了番号
  term* vtmroot;                // 項ツリーのルートポインター
  term** vtmtab;                // 項配列へのポインター
  int vntm;                     // 有効項数
  int order_expr(term* root, int idx); // 項ツリーをポインター配列に整列する
  int get_item(int& is, int& ie, const string& sstr) const; // 項目指定を取得する
  int get_expr(const string& pname, string& sstr, term** tmroot0, int ntm0, int level, int cont) const;
  // 式を取得する(式の構文解析)
  int check_term(int level, const string& sstr) const;
  // 文字列をチェックし終了文字列なら0以外、終了文字列以外なら0を返す
  idtype search_ope(int& level, const string& sstr) const;
  // 演算子とレベルをチェックして演算子タイプを返す
  int get_fact(const string& pname, string& sstr, term** fact, int ntm0) const;
  // 因子を取得する(因子の構文解析)
  int get_id(string& idstr, string& sstr) const;
  // 識別子(数字・項目指定・関数名)を小文字にして取り出す
  double id_todf(const string& idstr) const; // 識別子を数字に変換して返す
  idtype search_fn(const string& str) const; // 関数をサーチする
public:
  form() : vono(0, 0), vtmroot(nullptr), vtmtab(nullptr), vntm(0)
  // デフォールトコンストラクタ
  {}
  form(int os0, int oe0, term* tmroot0, int ntm0); // コンストラクタ
  form(const form& form0);      // コピーコンストラクタ
  virtual ~form();              // デストラクタ
  const form& operator=(const form& form0); // コピー代入演算子
  int os() const                // 出力項目の開始番号を返す
  { return vono.first; }
  int os(int os0)               // 出力項目の開始番号を設定する
  { return (vono.first = os0); }
  int oe() const                // 出力項目の終了番号を返す
  { return vono.second; }
  int oe(int oe0)               // 出力項目の終了番号を設定する
  { return (vono.second = oe0); }
  term* tmroot() const          // 項ツリーのルートポインターを返す
  { return vtmroot; }
  term* tmtab(int idx) const    // 項へのポインターを返す
  { return (idx < vntm) ? vtmtab[idx] : nullptr; }
  int ntm() const               // 有効項数を返す
  { return vntm; }
  int get(const string& pname, const string& oistr, const string& exstr);
  // 変換式を取り出す
};

form::form(int os0, int oe0, term* tmroot0, int ntm0) : vono(os0, oe0), vtmroot(tmroot0), vtmtab(0), vntm(ntm0)
// コンストラクタ
{
  if (vntm < 1) return;         // 有効項なし
  vtmtab = new term*[vntm];
  vntm = order_expr(vtmroot, 0);
}

form::form(const form& form0) : vono(form0.vono), vtmroot(0), vtmtab(0), vntm(form0.vntm)
// コピーコンストラクタ(オブジェクトを再帰的にコピーする)
{
  if (form0.vtmroot == nullptr) return;
  vtmroot = new term(*(form0.vtmroot));
  if (vntm < 1) return;         // 有効項なし
  vtmtab = new term*[vntm];
  vntm = order_expr(vtmroot, 0);
}

form::~form()                   // デストラクタ
{
  delete vtmroot;
  delete [] vtmtab;
}

const form& form::operator=(const form& form0) // コピー代入演算子
{
  if (&form0 == this) return *this; // 自分自身への代入を防ぐ
  vono = form0.vono;  vtmroot = nullptr;  vtmtab = nullptr;  vntm = form0.vntm;
  if (form0.vtmroot == nullptr) return *this;
  vtmroot = new term(*(form0.vtmroot));
  if (vntm < 1) return *this;   // 有効項なし
  vtmtab = new term*[vntm];
  vntm = order_expr(vtmroot, 0);
  return *this;
}

int form::order_expr(term* root, int idx) // 項ツリーをポインター配列に整列する
{
  if (root == nullptr) return idx;
  idx = order_expr(root->fact(), idx);
  idx = order_expr(root->pre(), idx);
  if (root->type() != idnop && root->type() != idfn) vtmtab[idx++] = root;
  return order_expr(root->next(), idx);
}

int form::get(const string& pname, const string& oistr, const string& exstr)
// 変換式を取得する
{
  if (oistr.size() > 0)
  {
    if (get_item(vono.first, vono.second, oistr) != 0) return disperr(pname, "出力項目指定が間違ってござる→", oistr);
  }
  string tstr(exstr);
  vntm = get_expr(pname, tstr, &vtmroot, 0, 0, 0);
  if (vntm < 0) return -1;      // 式の取得失敗
  if (vntm > 0)                 // 有効な項があるとき
  {
    vtmtab = new term*[vntm];
    vntm = order_expr(vtmroot, 0);
  }
  return 0;                     // 取得成功
}

int form::get_item(int& is, int& ie, const string& sstr) const
// 項目指定を取得する
{
  int tis = 1, tie = INT_MAX;
  if (getrange(tis, tie, sstr.substr(1)) < 1) return -1;
  if (tis < 1 || tis > tie) return -2;
  is = tis;  ie = tie;
  return 0;                     // 取得成功
}

int form::get_expr(const string& pname, string& sstr, term** tmroot0, int ntm0, int level, int cont) const
// 式を取得する(式の構文解析)
{
  string tstr(sstr);
  strtrm(tstr);
  if (check_term(level, tstr) != 0) return disperr(pname, "変換式が中途半端でござる→", sstr);
  int lexpr = (level == 0) ? 1 : level, lid = cont; // cont=0は式の最初
  idtype type = search_ope(lid, tstr); // 演算子とレベルのチェック
  if (type == idblank) return disperr(pname, "演算子が間違ってござる→", sstr);
  term* ptm = nullptr;
  if (*tmroot0 == nullptr)      // 式の最初の因子の取得
  {
    *tmroot0 = new term(type);  // 因子の前の演算子(level=0,1の演算子)
    if (type != idnop) { ++ntm0;  tstr.erase(0, 1); } // ヌル演算子以外を数える
    ptm = (*tmroot0)->fact();
    ntm0 = get_fact(pname, tstr, &ptm, ntm0);
    (*tmroot0)->fact(ptm);
    if (ntm0 < 0 || ptm->type() == idfntitle || ptm->type() == idfnitem) return ntm0;
    // エラーまたはtitle()またはitem()
  }
  ptm = *tmroot0;
  strtrm(tstr);
  while (ntm0 >= 0 && tstr.size() > 0 && check_term(level, tstr) == 0)
  {                             // 項の取得
    lid = 1;
    term* ptmp = nullptr;
    type = search_ope(lid, tstr); // 項の前の演算子(level=1-3の演算子)
    if (type == idblank)
    { ntm0 = disperr(pname, "二項演算子が間違ってござる→", sstr); }
    else if (lid > lexpr)       // より優先レベルの高い演算子
    {
      ptmp = ptm->pre();
      ntm0 = get_expr(pname, tstr, &ptmp, ntm0, lid, 1);
      ptm->pre(ptmp);
    }
    else if (lid == lexpr)      // 優先レベルの等しい演算子
    {
      ptmp = new term(type);
      if (type == idpower) ptm->pre(ptmp); // ベキ乗演算子は右結合性
      else ptm->next(ptmp);     // その他の演算子は左結合性
      ptm = ptmp;  ++ntm0;  tstr.erase(0, 1);
      ptmp = ptm->fact();
      ntm0 = get_fact(pname, tstr, &ptmp, ntm0);
      ptm->fact(ptmp);
    }
    strtrm(tstr);
  }
  sstr = tstr;                  // 文字列を更新して返す
  return ntm0;                  // 現在の有効項数を返す
}

int form::check_term(int level, const string& sstr) const
// 文字列をチェックし終了文字列なら0以外、終了文字列以外なら0を返す
{
  if (sstr.empty()) return 1;   // level=0-4での終了文字列
  if (level == 0) return 0;     // level=0での終了文字列以外
  char c = sstr[0];
  if (c == ')') return 1;       // level=1-4での終了文字列
  if (level == 1) return 0;     // level=1での終了文字列以外
  if (level < 4 && (c == '+' || c == '-')) return 2; // level=2-4での終了文字列
  if (level == 2) return 0;     // level=2での終了文字列以外
  if (c == '*' || c == '/' || c == '%') return 3; // level=3-4での終了文字列
  if (level == 3) return 0;     // level=3での終了文字列以外
  if (isspace(c) || c == '(' || c == '^') return 4; // level=4での終了文字列
  return 0;                     // level=4での終了文字列以外
}

idtype form::search_ope(int& level, const string& sstr) const
// 演算子とレベルをチェックして演算子タイプを返す
{
  if (sstr.empty()) return idblank;
  char c = sstr[0];
  if (level == 0)       // level=0(式の最初)は単項演算子だけをサーチする
  {
    level = 1;
    switch (c)
    {
    case '+' : return idplusop;
    case '-' : return idminusop;
    }
    return idnop;               // 演算子なし(ヌル演算子)
  }
  level = 1;
  if (c == '+') return idadd;
  if (c == '-') return idsub;
  level = 2;
  if (c == '*') return idmulti;
  if (c == '/') return iddivd;
  if (c == '%') return idmod;
  level = 3;
  if (c == '^') return idpower;
  return idblank;
}

int form::get_fact(const string& pname, string& sstr, term** fact, int ntm0) const
// 因子を取得する(因子の構文解析)
{
  string tstr(sstr);
  string idstr;
  if (get_id(idstr, tstr) < 0) return disperr(pname, "識別子が間違ってござる→", sstr);
  if (idstr.size() > 0 && (idstr[0] == '-' || idstr[0] == '+')) return disperr(pname, "識別子が間違ってござる→", sstr);
  double f = BLANKf;
  if (tstr[0] == '(')           // 関数
  {
    idtype type = search_fn(idstr);
    if (type == idblank) return disperr(pname, "こんな関数ないでござる→", sstr);
    *fact = new term(type);  ++ntm0;
    tstr.erase(0, 1);
    if (type == idfntitle || type == idfnitem) // title()・item()
    {
      if (ntm0 > 1) return disperr(pname, "式の途中にこの関数は置けないでござる→", sstr);
      string substr;
      strsize_t idx = getsubstr(substr, tstr, ')');
      if (idx == string::npos || idx >= tstr.size()) return disperr(pname, "関数の右括弧がないでござる→", tstr);
      if (substr.size() + 1 < tstr.size()) return disperr(pname, "この関数の後に式は指定できないでござる→", tstr);
      int n = (*fact)->get_sub(pname, substr);
      return (n < 0) ? n : ntm0; // n<0はエラー発生
    }
    term* ptm = (*fact)->fact();
    ntm0 = get_expr(pname, tstr, &ptm, ntm0, 1, 0);
    (*fact)->fact(ptm);
    if (ntm0 < 0) return ntm0;
    if (tstr.empty() || tstr[0] != ')') return disperr(pname, "右括弧[)]が足らんでござる→", sstr);
    if (type == idfnmax || type == idfnmean || type == idfnmin || type == idfnn || type == idfnsd || type == idfnsum || type == idfntmax || type == idfntmin)
    {                           // 統計関数
      if ((((*fact)->fact())->fact())->type() != iditem) return disperr(pname, "統計関数の引数は項目だけでござる→", sstr);
      (*fact)->is((((*fact)->fact())->fact())->is());
      (*fact)->ie((((*fact)->fact())->fact())->ie());
      (((*fact)->fact())->fact())->type(idnum);
      (((*fact)->fact())->fact())->f(0.0);
    }
    else if (type == idfnrand)  // 乱数関数の時、乱数の種をまく
    { srand((int)((((*fact)->fact())->fact())->f())); }
    tstr.erase(0, 1);
  }
  else if (idstr[0] == 'i')     // 項目指定
  {
    int is = 1, ie = INT_MAX;
    if (get_item(is, ie, idstr) != 0) return disperr(pname, "変換項目指定が間違ってござる→", sstr);
    *fact = new term(iditem, is, ie);  ++ntm0;
  }
  else                          // 数字
  {
    double f = id_todf(idstr);
    if (f == BLANKf) return disperr(pname, "識別子が何か変でござる→", sstr);
    *fact = new term(idnum, 0, 0, f);  ++ntm0;
  }
  sstr = tstr;                  // 文字列を更新して返す
  return ntm0;                  // 現在の有効項数を返す
}

int form::get_id(string& idstr, string& sstr) const
// 識別子(数字・項目指定・関数名)を小文字にして取り出す
{
  if (sstr.empty()) return -1;  // 取得できず
  idstr.erase();
  strtrm(sstr);                 // 識別子の前のスペース削除
  while (check_term(4, sstr) == 0)
  {
    char c = sstr[0];
    idstr += tolower(c);
    sstr.erase(0, 1);
  }
  return 0;                     // 取得成功
}

double form::id_todf(const string& idstr) const // 識別子を数字に変換して返す
{
  if (idstr.empty()) return BLANKf;
  if (idstr == "pi") return M_PI;
  if (idstr == "e") return exp(1.0);
  return strtodf(idstr);
}

idtype form::search_fn(const string& str) const // 関数をサーチする
{
  if (str.empty()) return idfn;
  if (str == "abs") return idfnabs;
  if (str == "cos") return idfncos;
  if (str == "exp") return idfnexp;
  if (str == "int") return idfnint;
  if (str == "item") return idfnitem;
  if (str == "ln") return idfnln;
  if (str == "log") return idfnlog;
  if (str == "max") return idfnmax;
  if (str == "mean") return idfnmean;
  if (str == "min") return idfnmin;
  if (str == "n") return idfnn;
  if (str == "rand") return idfnrand;
  if (str == "sd") return idfnsd;
  if (str == "sgn") return idfnsgn;
  if (str == "sin") return idfnsin;
  if (str == "sqrt") return idfnsqrt;
  if (str == "sum") return idfnsum;
  if (str == "tan") return idfntan;
  if (str == "title") return idfntitle;
  if (str == "tmax") return idfntmax;
  if (str == "tmin") return idfntmin;
  return idblank;
}

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  pairint vc;                   // 変換ケース番号
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
  int vprec;                    // 変換後データの有効数字
  list<form> vfmlst;            // 変換式リスト
  lstra vstralst;               // 層別情報リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int cs() const                // 変換開始ケース番号を返す
  { return vc.first; }
  int ce() const                // 変換終了ケース番号を返す
  { return vc.second; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  int prec() const              // 変換後データの有効数字を返す
  { return vprec; }
  const list<form>& fmlst() const // 変換式リストを返す
  { return vfmlst; }
  lstra& stralst()              // 層別情報リストを返す
  { return vstralst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  bool smatch(datac& cdata) const // 層別のチェック
  { return stramatch(vstralst, cdata); }
};

option::option(const char* c_pname) : vpath(), vpname(), vc(1, INT_MAX), vcd(true), vprec(0), vfmlst(), vid(true), vstralst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 変換ケース番号指定
  {
    if (getrange(vc.first, vc.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "cd")      // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0][0] == 'i')    // 変換式の指定
  {
    if (opt[1].empty())
    { err = badopt(vpname, optstr, USAGE); }
    else
    {
      form tfm;
      err = tfm.get(vpname, opt[0], opt[1]);
      if (!err) vfmlst.push_back(tfm);
    }
  }
  else if (opt[0] == "prec")    // 変換後データの有効数字指定
  {
    char* prt = nullptr;
    vprec = opt[1].empty() ? PRECdef : strtol(opt[1].c_str(), &prt, 0);
    if (vprec < 0 || PRECmax < vprec) err = badopt(vpname, optstr, USAGE, 0.0, (double)PRECmax);
  }
  else if (opt[0][0] == 'r' || opt[0][0] == 's') // 層別指定
  {
    stra tstra;
    char* prt = nullptr;
    int ino = strtol(opt[0].c_str() + 1, &prt, 0);
    if (tstra.get(opt[0][0], ino, opt[1]) == datanon) err = badopt(vpname, optstr, USAGE);
    if (!err) vstralst.push_back(tstra);
  }
  else if ((opt[0]).substr(0, 5) == "title") // データ名変更
  {
    if (opt[1].empty())
    {
      form tfm;
      err = tfm.get(vpname, opt[1], opt[0]);
      if (!err) vfmlst.push_back(tfm);
    }
    else
    { err = badopt(vpname, optstr, USAGE); }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (straok(vstralst, ni) == false) return disperr(vpname, "層別項目指定が間違っとるんよ！");
  int imax = (int)ni;
  list<form>::iterator itr = vfmlst.begin();
  while (itr != vfmlst.end())    // 変換後項目番号最大値をサーチする
  {
    if (itr->oe() == INT_MAX) itr->oe(ni);
    else if (itr->oe() > imax) imax = itr->oe();
    if (itr->os() > itr->oe()) return disperr(vpname, "出力項目指定が間違っておるでござる！");
    ++itr;
  }
  itr = vfmlst.begin();
  while (itr != vfmlst.end())
  {
    for (int i = 0; i < itr->ntm(); ++i)
    {
      term* ptm = itr->tmtab(i);
      if (ptm->ie() == INT_MAX) ptm->ie(ni);
      if (ptm->is() > ptm->ie() || ptm->ie() > imax) return disperr(vpname, "変換元項目指定が間違っておるでござる！");
      if ((itr->os() < itr->oe()) && (ptm->is() < ptm->ie()))
      {
        if (itr->oe() - itr->os() != ptm->ie() - ptm->is()) return disperr(vpname, "出力項目指定か変換元項目指定が間違っておるでござる！");
      }
    }
    ++itr;
  }
  return imax;                  // 変換後最大項目番号を返す
}

class convfclass                // convfメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id) const;
  // データ識別情報の変換と出力
  void put_convfcase(ostream& fout, const option& opt, const idd& id, datac& cdata) const;
  // ケースデータを変換して出力する
  double convf_data(const form& fm, const vector<double>& fdata, int idx) const;
  // 数字データを変換して返す
  double powf(double a, double x) const; // convf用ベキ乗関数
  double fnstat(idtype type, const vecdbl& fdata, int is, int ie) const;
  // 統計関数
  void psub(string& targstr, const list<carg>& carglst) const;
  // 部分文字列を置換して返す
public:
  convfclass() : fname()        // コンストラクタ
  {}
  ~convfclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込んで保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void convfclass::put_id(ostream& fout, const option& opt, idd& id) const
// データ識別情報の変換と出力
{
  string targstr;
  for (list<form>::const_iterator fmitr = (opt.fmlst()).begin(); fmitr != (opt.fmlst()).end(); ++fmitr)
  {
    term* titr = fmitr->tmtab(0);
    if (titr->type() == idfntitle) // データ名変換
    {
      targstr = id.title();
      psub(targstr, titr->carglst());
      id.set(targstr, id.ni());
    }
    else if (titr->type() == idfnitem) // 項目名変換
    {
      int ino = 0;
      list<carg>::const_iterator aitr = (titr->carglst()).begin();
      if (aitr->type() == sdata) targstr = aitr->str1();
      else if (aitr->is() == aitr->ie()) targstr = id.item(aitr->is() - 1);
      else ino = aitr->is();
      for (int i = fmitr->os(); i <= fmitr->oe(); ++i)
      {
        if (ino > 0)            // 変換対象が複数項目の場合
        {
          targstr = id.item(ino - 1);
          ++ino;
        }
        string tstr(targstr);
        psub(tstr, titr->carglst());
        if (aitr->type() == itemapp) tstr = id.item(i - 1) + tstr;
        id.assign_item(i - 1, tstr);
      }
    }
  }
  id.put(fout, formtag);
}

void convfclass::psub(string& targstr, const list<carg>& carglst) const
// 部分文字列を置換して返す
{
  for (list<carg>::const_iterator citr = carglst.begin(); citr != carglst.end(); ++citr)
  {
    int nrep = strregrep(targstr, citr->str1(), citr->str2(), citr->global());
    if (nrep < 0) throw runtime_error("文字列置換指定/" + citr->str1() + "/" + citr->str2() + "/が間違ってるよ！");
  }
}

void convfclass::put_convfcase(ostream& fout, const option& opt, const idd& id, datac& cdata) const
// ケースデータを変換して出力する
{
  vecdbl fdata(id.ni());
  vector<bool> fconv(id.ni(), false);
  for (size_t j = 0; j < id.ni(); ++j) fdata[j] = cdata.fdata(j);
  for (list<form>::const_iterator fmitr = (opt.fmlst()).begin(); fmitr != (opt.fmlst()).end(); ++fmitr)
  {
    if ((fmitr->tmtab(0))->type() != idfntitle && (fmitr->tmtab(0))->type() != idfnitem)
    {                           // title()とitem()は無視
      int sidx = fmitr->os() - 1, eidx = fmitr->oe() - 1;
      for (int j = sidx; j <= eidx; ++j)
      { fdata[j] = convf_data(*fmitr, fdata, j - sidx);  fconv[j] = true; }
      for (int j = sidx; j <= eidx; ++j) cdata.assign(j, fdata[j]);
    }
  }
  if (opt.prec() < 1)
  { cdata.put(fout); }
  else
  {
    int prec0 = fout.precision(); // 現在の浮動小数点精度を保存する
    fout.precision(opt.prec());   // 現在の浮動小数点精度を設定する
    if (fconv[0] == true && fdata[0] != BLANKf) fout << fdata[0];
    else fout << cdata.dstr(0);
    for (size_t j = 1; j < id.ni(); ++j)
    {
      if (fconv[j] == true && fdata[j] != BLANKf) fout << CHRdelm << fdata[j];
      else fout << CHRdelm << cdata.dstr(j);
    }
    fout.precision(prec0);
    fout << endl;
  }
}

double convfclass::convf_data(const form& fm, const vecdbl& fdata, int idx) const
// 数字データを変換して返す
{
  vecdbl regi(fm.ntm() + 1);    // 計算用スタックレジスター
  regi[0] = BLANKf;             // ダミー(番兵)
  double eps = DBL_EPSILON * 5.0; // 計算誤差
  int j = 0;
  for (int i = 0; i < fm.ntm(); ++i)
  {
    term* ptm = fm.tmtab(i);
    int k = 0;
    switch(ptm->type())
    {
    case idnum : ++j;           // 数字データをスタックにプッシュ
      regi[j] = ptm->f();
      break;
    case iditem : ++j;          // 項目データをスタックにプッシュ
      k = ((ptm->is() == ptm->ie()) ? ptm->is() : ptm->is() + idx) - 1;
      regi[j] = fdata[k];
      break;
    case idplusop : break;             // 単項[+]演算、何もしない!
    case idminusop : regi[j] = -regi[j]; // 単項[-]演算
      break;
    case idadd : regi[j - 1] += regi[j]; // 加算、スタックをポップ
      --j;
      break;
    case idsub : regi[j - 1] -= regi[j]; // 減算、スタックをポップ
      --j;
      break;
    case idmulti : regi[j - 1] *= regi[j]; // かけ算、スタックをポップ
      --j;
      break;
    case iddivd : regi[j - 1] = (regi[j] == 0.0) ? BLANKf : regi[j - 1] / regi[j];
      --j;                      // 除算、スタックをポップ
      break;
    case idmod : regi[j - 1] = (regi[j] == 0.0) ? BLANKf : fmod(regi[j - 1], regi[j]);
      --j;                      // 余り算、スタックをポップ
      break;
    case idpower : regi[j - 1] = powf(regi[j - 1], regi[j]);
      --j; // ベキ算、スタックをポップ
      break;
    case idfnabs : regi[j] = fabs(regi[j]); // 関数abs()
      break;
    case idfnsgn : regi[j] = sgn(regi[j]); //  関数sgn()
      break;
    case idfnint : regi[j] = long(regi[j]); //  関数int()
      break;
    case idfnsqrt : regi[j] = (regi[j] < 0.0) ? BLANKf : sqrt(regi[j]); //  関数sqrt()
      break;
    case idfnexp : regi[j] = exp(regi[j]); //  関数exp()
      break;
    case idfnlog : regi[j] = (regi[j] > 0.0) ? log10(regi[j]) : BLANKf; //  関数log()
      break;
    case idfnln : regi[j] = (regi[j] > 0.0) ? log(regi[j]) : BLANKf; //  関数ln()
      break;
    case idfnsin : regi[j] = sin(regi[j]); //  関数sin()
      break;
    case idfncos : regi[j] = cos(regi[j]); //  関数cos()
      break;
    case idfntan : regi[j] = tan(regi[j]); //  関数tan()
      break;
    case idfnmax :              // 統計関数
    case idfnmean :
    case idfnmin :
    case idfnn :
    case idfnsd :
    case idfnsum :
    case idfntmax :
    case idfntmin : regi[j] = fnstat(ptm->type(), fdata, ptm->is(), ptm->ie());
      break;
    case idfnrand : regi[j] = rand(); // 乱数発生関数rand()
      break;
    default : regi[j] = BLANKf; // 変換指定なし
    }
    if (j < 1 || regi[j] == BLANKf || regi[j] == DBL_MAX) break;
    else if (fabs(regi[j]) < eps) regi[j] = 0.0;
  }
  if (j != 1) return BLANKf;
  return regi[1];
}

double convfclass::powf(double a, double x) const // convf用ベキ乗関数
{
  double dm = 0.0;
  if (a > 0.0) return pow(a, x);
  if (a < 0.0) return (modf(x, &dm) == 0.0) ? pow(a, x) : BLANKf;
  if (x == 0.0) return 1.0;
  return (x > 0.0) ? pow(a, x) : BLANKf;
}

double convfclass::fnstat(idtype type, const vector<double>& fdata, int is, int ie) const
// 統計関数
{
  if (is < 1 || ie < 1) return BLANKf;
  statv<double> s;
  double fmin = DBL_MAX, fmax = -DBL_MAX;
  int imin = -1, imax = -1;
  for (int i = is - 1; i < ie; ++i)
  {
    if (fdata[i] != BLANKf)
    {
      s.sigma(fdata[i]);
      if (fdata[i] < fmin) { fmin = fdata[i];  imin = i; }
      if (fdata[i] > fmax) { fmax = fdata[i];  imax = i; }
    }
  }
  int n = s.n();
  if (type == idfnn) return (double)n;
  if (n < 1) return BLANKf;
  switch (type)
  {
  case idfnmax : return fmax;
  case idfnmean : return s.mean();
  case idfnmin : return fmin;
  case idfnsd : return s.sd();
  case idfnsum : return s.sum();
  case idfntmax : return (double)(imax - is + 1);
  case idfntmin : return (double)(imin - is + 1);
  }
  return BLANKf;
}

int convfclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込んで保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  int imax = opt.check(id.ni()); // オプションと変換後最大項目のチェック
  if (imax < 0) throw option_error("実行時オプションエラー！");
  if (imax > (int)(id.ni())) id.set(id.title(), imax); // 項目数を増やす
  return nc;                    // 読み込んだケース数を返す
}

void convfclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return;
  int cno = 1;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++cno)
  {
    ditr->resize(id.ni());
    if (opt.cs() <= cno && cno <= opt.ce() && opt.smatch(*ditr)) put_convfcase(cout, opt, id, *ditr);
    else ditr->put(cout);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  convfclass convf;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) convf.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { convf.get(argv[i], opt, id, datalst);  fok = false; }
    }
    convf.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
