// convi.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 09:44:41 tac>
//
// dansのデータファイルを入力し、データの項目単位の変換を行う。
//
// 使用ライブラリ:libfun.so, libdans.so
// コンパイル法 :g++ -O -s -W convi convi.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "convi  Version 7.41          Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:convi [\"-<変換式>\" …] [<オプション> …] [<データファイル> …]",
  "機能:指定された変換式に従って項目単位のデータ変換を行う。",
  "変換式:i[<変換後データ出力開始項目番号>][-[<終了項目番号>]]=<文字列関数>",
  "識別情報関数:item(),title()",
  "データ変換関数:comp(),comps(),cordev(),corline(),cormid(),cum(),dd()",
  "  dev(),dif(),mean(),norm(),rand(), randn(), sd()",
  "<オプション>",
  "  -c=[<開始ケース番号>][-[<終了ケース番号>]]:変換ケース番号の指定。",
  "  -cd[={-|+}]:ケースデータの出力指定なんよ。(-cd:-cd=+)",
  "  -f=<ファイル名>:変換式やオプションを指定したファイルから読み込む。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示する。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定。(-id:-id=+)",
  "  -prec[=<有効数字>]:変換後データの有効数字の指定でござる。(0≦有効数字≦32)",
  "  -r<層別項目番号>=[<実数下限>][-[<実数上限>]]:実数層別項目指定",
  "  -s<層別項目番号>=<文字列>[,<文字列>,…]:文字層別項目指定",
  "    指定した実数範囲内または文字列と一致するケースだけ変換する。"
};

constexpr int PRECdef = 16;     // 省略時変換後データの有効数字
constexpr int PRECmax = 32;     // 変換後データの有効数字最大値
constexpr double Mab40r = 24.7242840716733; // Aβ1-40/Aβ1-42の母平均推定値
constexpr double SDab40r = 4.31242636217402; // Aβ1-40/Aβ1-42の母標準偏差推定値
constexpr double Mappr = 0.773838808675519; // APP669-711/Aβ1-42の母平均推定値
constexpr double SDappr = 0.190788767572489; // APP669-711/Aβ1-42の母標準偏差推定値

using vecstatv = vector<statv<double>>; // 統計量ベクトルの短縮名定義

enum fntype                     // 関数の種類
{
  fnblank,                      // 指定なし
  fncomp,                       // composite biomarker変換
  fncomps,                      // 母平均と母SDによるcomposite biomarker変換
  fncordev,                     // 偏差による補正
  fncorline,                    // 折れ線による補正
  fncormid,                     // 2点の平均値による補正
  fncum,                        // 累積変換関数
  fndd,                         // 中心差分変換関数
  fndev,                        // 偏差変換関数
  fndif,                        // 変化量変換関数
  fnitem,                       // 項目名変換関数
  fnmean,                       // 平均値変換関数
  fnnorm,                       // 標準化変換関数
  fnrand,                       // 乱数出力関数
  fnrandn,                      // 正規乱数出力関数
  fnsd,                         // 標準偏差変換関数
  fntitle,                      // データ名変換関数
  fnetc                         // その他の関数
};

enum cargtype                   // 引数の種類
{
  cargblank,                    // 引数なし
  itemsub,                      // 項目
  itemapp,                      // 項目名追加
  sdata,                        // 文字列データ
  cargetc                       // その他
};

class carg                      // 引数クラス
{
private:
  cargtype vtype;               // 引数の種類
  pairint irange;               // 変換項目の開始・終了番号
  string vstr1;                 // 第1文字列(変換対象文字列等)
  string vstr2;                 // 第2文字列(変換後文字列等)
  bool vglobal;                 // グローバル置換フラッグ
  char vctype;                  // 変換指定型('i','r'等)
  pairdbl frange;               // 実数の上・下限(範囲指定等)
public:
  carg() : vtype(cargblank), irange(0, 0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // デフォールトコンストラクタ
  {}
  explicit carg(cargtype type0, int is0 = 0, int ie0 = 0) : vtype(type0), irange(is0, ie0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // コンストラクタ
  {}
  carg(cargtype type0, int is0, int ie0, const string& str10, const string& str20, bool global0 = false, char ctype0 = ' ', double low0 = BLANKf, double up0 = BLANKf) : vtype(type0), irange(is0, ie0), vstr1(str10), vstr2(str20), vglobal(global0), vctype(ctype0), frange(low0, up0)
  // コンストラクタ
  {}
  carg(const carg& carg0) : vtype(carg0.vtype), irange(carg0.irange), vstr1(carg0.vstr1), vstr2(carg0.vstr2), vglobal(carg0.vglobal), vctype(carg0.vctype), frange(carg0.frange)
  // コピーコンストラクタ
  {}
  virtual ~carg()               // デストラクタ
  {}
  const carg& operator=(const carg& carg0); // コピー代入演算子
  cargtype type() const                     // 引数の種類を返す
  { return vtype; }
  cargtype type(cargtype type0) // 引数の種類を設定する
  { return (vtype = type0); }
  int is() const                // 変換項目の開始番号を返す
  { return irange.first; }
  int is(int is0)               // 変換項目の開始番号を設定する
  { return (irange.first = is0); }
  int ie() const                // 変換項目の終了番号を返す
  { return irange.second; }
  int ie(int ie0)               // 変換項目の終了番号を設定する
  { return (irange.second = ie0); }
  const string& str1() const    // 文字列1を返す
  { return vstr1; }
  const string& str1(const string& str10) // 文字列1を設定する
  { return (vstr1 = str10); }
  const string& str2() const    // 文字列2を返す
  { return vstr2; }
  const string& str2(const string& str20) // 文字列2を設定する
  { return (vstr2 = str20); }
  bool global() const           // グローバル置換フラッグを返す
  { return vglobal; }
  bool global(bool global0)     // グローバル置換フラッグを設定する
  { return (vglobal = global0); }
  char ctype() const            // 変換指定型を返す
  { return vctype; }
  char ctype(char ctype0)       // 変換指定型を設定する
  { return (vctype = ctype0); }
  double low() const            // 実数の下限を返す
  { return frange.first; }
  double low(double low0)       // 実数の下限を設定する
  { return (frange.first = low0); }
  double up() const             // 実数の上限を返す
  { return frange.second; }
  double up(double up0)         // 実数の上限を設定する
  { return (frange.second = up0); }
};

const carg& carg::operator=(const carg& carg0) // コピー代入演算子
{
  if (&carg0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = carg0.vtype;  irange = carg0.irange;
  vstr1 = carg0.vstr1;  vstr2 = carg0.vstr2;  vglobal = carg0.vglobal;
  vctype = carg0.vctype;  frange = carg0.frange;
  return *this;
}

using lcarg = list<carg>;       // 引数リストの短名定義

class expr                      // 変換式クラス
{
private:
  fntype vtype;                 // 関数の種類
  pairint irange;               // 出力項目の開始・終了番号
  lcarg vcarglst;               // 引数リスト
  cargtype get_target(int& is, int& ie, string& targstr, const string& sstr) const;
  // 変換対象項目または変換対象文字列を取得する
  int get_expr(const string& pname, const string& sstr); // 変換式を取得する
  fntype search_fn(const string& str) const; // 関数をサーチする
  int get_title(const string& pname, const string& cargstr);
  // title()関数の引数を取得する
  int get_item(const string& pname, const string& cargstr);
  // item()関数の引数を取得する
  int get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr) const;
  // 置換対象文字列と置換後文字列とグローバル指定を取得する
  int get_comp(const string& pname, const string& cargstr);
  // comp()関数の引数を取得する
  int get_cor(const string& pname, const string& cargstr);
  // cor.*()関数の引数を取得する
  int get_dd(const string& pname, const string& cargstr);
  // dd()関数の引数を取得する
  int get_stat(const string& pname, const string& cargstr);
  // stat系関数の引数を取得する
public:
  expr() : vtype(fnblank), irange(0, 0), vcarglst() // デフォールトコンストラクタ
  {}
  expr(fntype type0, int os0, int oe0, const lcarg& carglst0) : vtype(type0), irange(os0, oe0), vcarglst(carglst0)
  // コンストラクタ
  {}
  expr(const expr& expr0) : vtype(expr0.vtype), irange(expr0.irange), vcarglst(expr0.vcarglst)
  // コピーコンストラクタ
  {}
  virtual ~expr()               // デストラクタ
  {}
  const expr& operator=(const expr& expr0); // コピー代入演算子
  fntype type() const                       // 関数の種類を返す
  { return vtype; }
  fntype type(fntype type0)     // 関数の種類を設定する
  { return (vtype = type0); }
  int os() const                // 出力項目の開始番号を返す
  { return irange.first; }
  int os(int os0)               // 出力項目の開始番号を設定する
  { return (irange.first = os0); }
  int oe() const                // 出力項目の終了番号を返す
  { return irange.second; }
  int oe(int oe0)               // 出力項目の終了番号を設定する
  { return (irange.second = oe0); }
  lcarg& carglst()              // 引数リストを返す
  { return vcarglst; }
  int get(const string& pname, const string& oistr, const string& exstr);
  // 変換式を取り出す
};

const expr& expr::operator=(const expr& expr0) // コピー代入演算子
{
  if (&expr0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = expr0.vtype;  irange = expr0.irange;  vcarglst = expr0.vcarglst;
  return *this;
}

int expr::get(const string& pname, const string& oistr, const string& exstr)
// 変換式を取得する
{
  if (oistr.size() > 0)
  {
    string tstr;
    if (get_target(irange.first, irange.second, tstr, oistr) == cargblank) return disperr(pname, "出力項目指定が間違い→", oistr);
  }
  int err = get_expr(pname, exstr);
  if (err) return err;
  if (vtype == fntitle && oistr.size() > 0) return disperr(pname, "title()は単独で使用する→", exstr);
  return 0;                     // 取得成功
}

cargtype expr::get_target(int& is, int& ie, string& targstr, const string& sstr) const
// 変換対象項目または変換対象文字列を取得する
{
  targstr = sstr;
  strtrm(targstr);
  if (targstr.size() < 1) return cargblank;
  is = 0;  ie = 0;
  if (targstr[0] == 'i')        // 項目指定の場合
  {
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(1)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemsub;
  }
  if (targstr.size() > 1 && targstr[0] == '+' && targstr[1] == 'i')
  {                             // 項目名追加の場合
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(2)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemapp;
  }
  sstrtostr(targstr);           // 特殊表現文字を変換し、""を削除する
  return sdata;                 // 文字列指定の場合
}

int expr::get_expr(const string& pname, const string& sstr) // 変換式を取得する
{
  string tstr(sstr), substr;
  strsize_t idx = getsubstr(substr, tstr, '(');
  if (idx == string::npos || idx >= tstr.size()) return disperr(pname, "左括弧[(]がない→", sstr);
  strtrm(substr);
  strlower(substr);             // 関数名を小文字にする
  if ((vtype = search_fn(substr)) == fnblank) return disperr(pname, "こんな関数はない→", sstr);
  tstr.erase(0, idx + 1);
  idx = getsubstr(substr, tstr, ')');
  if (idx == string::npos || idx >= tstr.size()) return disperr(pname, "右括弧[)]がない→", sstr);
  strtrm(substr);
  int n = 0, is = 0, ie = 0;
  cargtype atype = cargblank;
  string targstr;
  carg tag;
  bool subf = true;
  switch (vtype)
  {
  case fntitle : n = get_title(pname, substr);
    break;
  case fnitem : n = get_item(pname, substr);
    break;
  case fncomp :
  case fncomps : n = get_comp(pname, substr);
    break;
  case fncordev :
  case fncorline :
  case fncormid : n = get_cor(pname, substr);
    break;
  case fndd : n = get_dd(pname, substr);
    break;
  case fnrand :
  case fnrandn : n = 0;
    if (substr.size() > 0) n = disperr(pname, "この関数に引数はいらないんよ→", sstr);
    break;
  case fncum :
  case fndev :
  case fndif :
  case fnmean :
  case fnnorm :
  case fnsd : n = get_stat(pname, substr);
    break;
  default : atype = get_target(is, ie, targstr, substr); // その他の引数
    if (atype == cargblank)
    { n = disperr(pname, "変換指定項目が間違っとるんよ→", sstr); }
    else
    {
      tag.type(atype);  tag.is(is);  tag.ie(ie);  tag.str1(targstr);
      vcarglst.push_back(tag);
    }
  }
  if (n < 0) vtype = fnblank;
  return (n < 0) ? n : 0;       // 取得成功は0を返す
}

fntype expr::search_fn(const string& str) const // 関数をサーチする
{
  if (str == "comp") return fncomp;
  if (str == "comps") return fncomps;
  if (str == "cordev") return fncordev;
  if (str == "corline") return fncorline;
  if (str == "cormid") return fncormid;
  if (str == "cum") return fncum;
  if (str == "dd") return fndd;
  if (str == "dev") return fndev;
  if (str == "dif") return fndif;
  if (str == "item") return fnitem;
  if (str == "norm") return fnnorm;
  if (str == "mean") return fnmean;
  if (str == "sd") return fnsd;
  if (str == "rand") return fnrand;
  if (str == "randn") return fnrandn;
  if (str == "title") return fntitle;
  return fnblank;
}

int expr::get_title(const string& pname, const string& cargstr)
// title()関数の引数を取得する(置換文字列指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "変換指定が間違っとるんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  string targstr, pattern, repstr;
  bool global = false;
  int err = 0, n = 0;
  while (!err && itr != cargstrlst.end())
  {
    if (get_pattern(targstr, pattern, repstr, global, *itr) < 0)
    { err = disperr(pname, "変換指定が間違っとるんよ→", cargstr); }
    else
    {
      carg tag(sdata, 0, 0, pattern, repstr, global);
      vcarglst.push_back(tag);
      ++itr;
    }
  }
  return err ? err : n;         // エラーでなければ置換数を返す
}

int expr::get_item(const string& pname, const string& cargstr)
// item()関数の引数を取得する
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "変換内容が間違っとるんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  string targstr, pattern, repstr;
  bool global = false;
  if (get_pattern(targstr, pattern, repstr, global, *itr) < 0) return disperr(pname, "最初の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string tstr;
  cargtype atype = get_target(is, ie, tstr, targstr);
  if (atype != itemsub && atype != itemapp) return disperr(pname, "この関数の変換対象指定は項目だけなんよ→", cargstr);
  carg tag(atype, is, ie, pattern, repstr, global);
  vcarglst.push_back(tag);
  ++itr;
  int err = 0, n = 1;
  while (!err && itr != cargstrlst.end())
  {
    if (get_pattern(targstr, pattern, repstr, global, *itr) < 0)
    { err = disperr(pname, "2番目以後の変換指定が間違っとるんよ→", cargstr); }
    else
    {
      tag.str1(pattern);  tag.str2(repstr);  tag.global(global);
      vcarglst.push_back(tag);
      ++itr;
    }
  }
  return err ? err : n;         // エラーでなければ置換数を返す
}

int expr::get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr) const
// 置換対象文字列と置換後文字列とグローバル指定を取得する
{
  vecstr cargstrvec;
  if (strsplit(cargstrvec, cargstr, CHRstr) != 4) return -1;
  targstr = cargstrvec[0];  pattern = cargstrvec[1];  repstr = cargstrvec[2];
  sstrtostr(pattern);  sstrtostr(repstr); // 特殊表現文字列を変換し""を削除する
  global = (cargstrvec[3] == "g") ? true : false;
  return 0;
}

int expr::get_comp(const string& pname, const string& cargstr)
// comp()関数の引数を取得する
{
  vecstr cargstrvec; 
 if (strsplit(cargstrvec, cargstr, CHRdelm) != 3) return disperr(pname, "comp()の引数は3つ→", cargstr);
  int is = 0, ie = 0;
  string tstr;
  cargtype atype = get_target(is, ie, tstr, cargstrvec[0]);
  if (atype != itemsub || is != ie) return disperr(pname, "comp()の最初の変換項目指定が不適切→", cargstr);
  carg tcarg(atype, is, ie);
  vcarglst.push_back(tcarg);
  atype = get_target(is, ie, tstr, cargstrvec[1]);
  if (atype != itemsub || is != ie) return disperr(pname, "comp()の2番目の変換項目指定が不適切→", cargstr);
  tcarg.type(atype);  tcarg.is(is);  tcarg.ie(ie);
  vcarglst.push_back(tcarg);
  atype = get_target(is, ie, tstr, cargstrvec[2]);
  if (atype == itemsub)
  {
    if (is != ie) return disperr(pname, "comp()の3番目の変換項目指定が不適切→", cargstr);
    tcarg.type(atype);  tcarg.is(is);  tcarg.ie(ie);
  }
  else if (atype == sdata)
  {
    if (tstr != "ratio") return disperr(pname, "comp()の3番目の変換指定文字列が不適切→", cargstr);
    tcarg.type(atype);  tcarg.str1(tstr);
  }
  else
  { return disperr(pname, "comp()の3番目の変換指定が不適切→", cargstr); }
  vcarglst.push_back(tcarg);
  return vcarglst.size();       // 引数リストの数を返す
}

int expr::get_cor(const string& pname, const string& cargstr)
// cor.*()関数の引数を取得する
{
  vecstr cargstrvec;
  int ncarg = strsplit(cargstrvec, cargstr, CHRdelm);
  if (ncarg < 3 || 4 < ncarg) return disperr(pname, "この関数の引数は3つか4つ→", cargstr);
  int is = 0, ie = 0;
  string tstr(cargstr);
  vecstr substrvec;
  cargtype atype = cargblank;
  double fdata = BLANKf;
  if (strsplit(substrvec, cargstrvec[0], '=') == 2) // 基準値指定あり
  {
    atype = get_target(is, ie, tstr, substrvec[0]);
    if (atype != itemsub || is != ie) return disperr(pname, "最初の層別項目指定が不適切→", cargstr);
    if (substrvec[1] == "total") // 層別項目の全体平均値を基準値にする
    { atype = cargetc; }
    else                        // 基準値指定
    {
      atype = sdata;
      fdata = strtodf(substrvec[1]);
      if (fdata == BLANKf) return disperr(pname, "最初の基準値指定が不適切→", cargstr);
    }
  }
  else
  {
    atype = get_target(is, ie, tstr, cargstrvec[0]);
    if (atype != itemsub || is != ie) return disperr(pname, "最初の層別項目指定が不適切→", cargstr);
  }
  carg tcarg(atype, is, ie, tstr, cargstr, false, ' ', fdata, fdata);
  vcarglst.push_back(tcarg);
  substrvec.clear();
  if (strsplit(substrvec, cargstrvec[1], '=') < 2) return disperr(pname, "2番目のID指定が不適切→", cargstr);
  atype = get_target(is, ie, tstr, substrvec[0]);
  if (atype != itemsub || is != ie) return disperr(pname, "2番目のID項目指定が不適切→", cargstr);
  string codestr(substrvec[1]);
  sstrtostr(codestr);           // 特殊表現文字を変換
  if (codestr[0] != CHRhead) codestr = CHRhead + codestr; // 完全一致にする
  if (codestr[codestr.size() -1] != CHRtail) codestr += CHRtail;
  tcarg.type(atype);  tcarg.is(is);  tcarg.ie(ie);
  tcarg.str1(codestr);  tcarg.str2(substrvec[0]);
  vcarglst.push_back(tcarg);
  atype = get_target(is, ie, tstr, cargstrvec[2]);
  if (atype != itemsub) return disperr(pname, "3番目の変換項目指定が不適切→", cargstr);
  if (ncarg > 3)
  {
    if (cargstrvec[3] != "i" && cargstrvec[3] != "r") return disperr(pname, "4番目の補正法指定はiまたはrのみ→", cargstr);
    tcarg.ctype(cargstrvec[3][0]);
  }
  tcarg.type(atype);  tcarg.is(is);  tcarg.ie(ie);
  vcarglst.push_back(tcarg);
  return ncarg;                 // 成功の場合は引数リスト数を返す
}

int expr::get_dd(const string& pname, const string& cargstr)
// dd()関数の引数を取得する
{
  vecstr cargstrvec; 
 if (strsplit(cargstrvec, cargstr, CHRdelm) != 2) return disperr(pname, "dd()の引数は2つ→", cargstr);
  int is = 0, ie = 0;
  string tstr;
  cargtype atype = get_target(is, ie, tstr, cargstrvec[0]);
  if (atype != itemsub) return disperr(pname, "dd()の最初の変換元項目指定が不適切→", cargstr);
  carg tcarg(atype, is, ie);
  vcarglst.push_back(tcarg);
  atype = get_target(is, ie, tstr, cargstrvec[1]);
  if (atype != itemsub || is != ie) return disperr(pname, "dd()の2番目の説明変数項目指定が不適切→", cargstr);
  tcarg.type(atype);  tcarg.is(is);  tcarg.ie(ie);
  vcarglst.push_back(tcarg);
  return vcarglst.size();       // 引数リストの数を返す
}

int expr::get_stat(const string& pname, const string& cargstr)
// stat系関数の引数を取得する
{
  int is = 0, ie = 0;
  string tstr;
  cargtype atype = get_target(is, ie, tstr, cargstr);
  if (atype != itemsub) return disperr(pname, "この関数の変換対象指定は項目だけ→", cargstr);
  carg tcarg(atype, is, ie);
  vcarglst.push_back(tcarg);
  return 1;                     // 変換指定は1個
}

using lexpr = list<expr>;       // 変換式リストの短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  pairint crange;               // 変換ケース番号
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
  int vprec;                    // 変換後データの有効数字
  lexpr vexlst;                 // 変換式リスト
  lstra vstralst;               // 層別情報リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int cs() const                // 変換開始ケース番号を返す
  { return crange.first; }
  int ce() const                // 変換終了ケース番号を返す
  { return crange.second; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  int prec() const              // 変換後データの有効数字を返す
  { return vprec; }
  lexpr& exlst()                // 変換式リストを返す
  { return vexlst; }
  lstra& stralst()              // 層別情報リストを返す
  { return vstralst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int ni);              // オプションのチェックとセット
  bool smatch(datac& cdata) const // 層別のチェック
  { return stramatch(vstralst, cdata); }
};

option::option(const char* c_pname) : vpath(), vpname(), crange(1, INT_MAX), vcd(true), vprec(0), vexlst(), vid(true), vstralst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 変換ケース番号指定
  {
    if (getrange(crange.first, crange.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "cd")      // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0][0] == 'i')    // 変換式の指定
  {
    if (opt[1].empty())
    { err = badopt(vpname, optstr, USAGE); }
    else
    {
      expr tex;
      err = tex.get(vpname, opt[0], opt[1]);
      if (!err) vexlst.push_back(tex);
    }
  }
  else if (opt[0] == "prec")    // 変換後データの有効数字指定
  {
    char* prt = nullptr;
    vprec = opt[1].empty() ? PRECdef : strtol(opt[1].c_str(), &prt, 0);
    if (vprec < 0 || PRECmax < vprec) err = badopt(vpname, optstr, USAGE, 0.0, double(PRECmax));
  }
  else if (opt[0][0] == 'r' || opt[0][0] == 's') // 層別指定
  {
    stra tstra;
    char* prt = nullptr;
    int ino = strtol(opt[0].c_str() + 1, &prt, 0);
    if (tstra.get(opt[0][0], ino, opt[1]) == datanon) err = badopt(vpname, optstr, USAGE);
    if (!err) vstralst.push_back(tstra);
  }
  else if ((opt[0]).substr(0, 5) == "title") // データ名変更
  {
    if (opt[1].empty())
    {
      expr tex;
      err = tex.get(vpname, opt[1], opt[0]);
      if (!err) vexlst.push_back(tex);
    }
    else
    { err = badopt(vpname, optstr, USAGE); }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int ni)       // オプションのチェックとセット
{
  if (straok(vstralst, ni) == false) return disperr(vpname, "層別項目指定が間違っとるんよ！");
  bool err = false;
  int imax = ni;
  lexpr::iterator itr = vexlst.begin();
  while (err == false && itr != vexlst.end()) // 変換後項目番号最大値をサーチする
  {
    if (itr->oe() == INT_MAX) itr->oe(ni);
    else if (itr->oe() > imax) imax = itr->oe();
    if (itr->os() > itr->oe()) err = true;
    else ++itr;
  }
  if (err) return disperr(vpname, "出力項目指定が間違っとるんよ！");
  itr = vexlst.begin();
  while (!err && itr != vexlst.end())
  {
    lcarg::iterator aitr = (itr->carglst()).begin();
    while (!err && aitr != (itr->carglst()).end())
    {
      if (aitr->ie() == INT_MAX) aitr->ie(ni);
      if (aitr->is() > aitr->ie() || aitr->ie() > imax)
      { err = true; }
      else if ((itr->os() < itr->oe()) && (aitr->is() < aitr->ie()))
      {                         // 出力指定項目が複数の時
        if (itr->oe() - itr->os() != aitr->ie() - aitr->is()) err = true;
        else ++aitr;
      }
      else
      { ++aitr; }
    }
    if (!err) ++itr;
  }
  if (err) return disperr(vpname, "出力項目指定か変換元項目指定が間違っとるんよ！");
  return imax;                  // 変換後最大項目番号を返す
}
  
class conviclass                // conviメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, option& opt, idd& id) const;
  // データ識別情報の変換と出力
  void put_id(ostream& fout) const; // データ識別情報の変換と出力
  void conv_data(option& opt, const idd& id, ldata& datalst, vecbool& fconv) const;
  // データを変換する
  void psub(string& targstr, const lcarg& carglst) const;
  // 部分文字列を置換して返す
  void set_statvec(lcarg::iterator citr, vecstatv& statvec, map<string, vecstatv>& vstatmap) const;
  // 基準値をセットする
  int get_qc(const option& opt, ldata& datalst, datac& qcdata0, datac& qcdata1, int& nd, string qccode, ldata_itr ditr, int sidx, int ididx, int cno, int didx) const;
  // QCデータを取得して現在のデータインデックスを返す
  void cal_convi(lexpr::iterator eitr, vecstatv& statvec, map<string, vecstatv>& vstatmap, datac& cdata) const;
  // 項目データ変換用にデータを集計する
  void cal_cor(lcarg::iterator citr, vecstatv& statvec, map<string, vecstatv>& vstatmap, datac& cdata) const;
  // cor.*()用にデータを集計する
  void cal_comp(lcarg::iterator citr, vecstatv& statvec, datac& cdata) const;
  // comp()用にデータを集計する
  void conv_case(lexpr::iterator eitr, vecstatv& statvec, map<string, vecstatv>& vstatmap, bool firstd, bool lastd, datac& cdata0, datac& cdata, datac& cdata1, datac& qcdata0, datac& qcdata1, int didx, int nd) const;
  // ケースデータを変換して出力する
  void conv_cordev(datac& cdata, vecstatv& statvec, map<string, vecstatv>& vstatmap, map<string, vecstatv>::iterator mitr, int sidx, int idx, int oidx, int kidx, char cormeth) const;
  // ケースデータをcordev()変換する
  void conv_corlm(datac& cdata, datac& qcdata0, datac& qcdata1, vecstatv& statvec, map<string, vecstatv>& vstatmap, map<string, vecstatv>::iterator mitr, int sidx, int idx, int oidx, int kidx, int didx, int nd, cargtype catype, char cormeth, fntype etype) const;
  // ケースデータをcorline()・cormid()変換する
  void conv_comp(datac& cdata, fntype ftype, vecstatv& statvec, lcarg::iterator citr, int isidx, int oidx) const;
  // ケースデータをcomp()変換する
  void conv_rand(datac& cdata, vecstatv& statvec, int oidx, int kidx) const;
  // 乱数を出力する
  void conv_dd(datac& cdata, int isidx, int oidx) const;
  // ケースデータをdd()変換する
  void conv_stat(bool firstd, bool lastd, datac& cdata0, datac& cdata, datac& cdata1, vecstatv& statvec, lexpr::iterator eitr, int idx, int oidx, int kidx, int xidx) const;
  // ケースデータをstat系関数変換する
  double cal_dd(datac& cdata0, datac& cdata, datac& cdata1, int idx, int xidx) const;
  // 中心差分変換値を返す
public:
  conviclass() : fname()        // コンストラクタ
  {}
  ~conviclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込んで保存する
  void put(option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void conviclass::put_id(ostream& fout, option& opt, idd& id) const
// データ識別情報の変換と出力
{
  string targstr;
  for (lexpr::iterator exitr = (opt.exlst()).begin(); exitr != (opt.exlst().end()); ++exitr)
  {
    if (exitr->type() == fntitle) // データ名変換
    {
      targstr = id.title();
      psub(targstr, exitr->carglst());
      id.set(targstr, id.ni());
    }
    else if (exitr->type() == fnitem) // 項目名変換
    {
      int ino = 0;
      lcarg::const_iterator aitr = (exitr->carglst()).begin();
      if (aitr->type() == sdata) targstr = aitr->str1();
      else if (aitr->is() == aitr->ie()) targstr = id.item(aitr->is() - 1);
      else ino = aitr->is();
      for (int i = exitr->os(); i <= exitr->oe(); ++i)
      {
        if (ino > 0)            // 変換対象が複数項目の場合
        {
          targstr = id.item(ino - 1);
          ++ino;
        }
        string tstr(targstr);
        psub(tstr, exitr->carglst());
        if (aitr->type() == itemapp) tstr = id.item(i - 1) + tstr;
        id.assign_item(i - 1, tstr);
      }
    }
  }
  id.put(fout, formtag);
}

void conviclass::conv_data(option& opt, const idd& id, ldata& datalst, vecbool& fconv) const
// データを変換する
{
  for (lexpr::iterator eitr = (opt.exlst()).begin(); eitr != (opt.exlst()).end(); ++eitr)
  {
    fntype etype = eitr->type();
    if (etype != fntitle && etype != fnitem)
    {
      vecstatv statvec(2);
      map<string, vecstatv> vstatmap;
      lcarg::iterator citr = (eitr->carglst()).begin();
      int isidx = citr->is() - 1, ieidx = citr->ie() - 1;
      int sidx = 0, ididx = 0, didx = 0, nd = 0, nii = 2;
      string qccode;
      datac qcdata0(id.ni()), qcdata1(id.ni());
      bool firstd = true, lastd = false;
      ldata_itr ditr0 = datalst.begin();
      switch (etype)
      {
      case fncum :
      case fndif :
      case fndd :
      case fncomp :
      case fncomps : break;     // 何もしない
      case fnrand :
      case fnrandn : srand(eitr->os()); // 乱数の種をまく
        nii = eitr->oe() - eitr->os() + 1; // 出力項目数
        break;
      case fncordev :
      case fncorline :
      case fncormid : sidx = isidx;
        ++citr;
        ididx = citr->is() - 1; // ID項目のインデックス
        qccode = citr->str1();  // QCデータのコード
        ++citr;
        nii = citr->ie() - citr->is() + 1; // 集計項目数
        break;
      default : nii = ieidx - isidx + 1; // 集計項目数
      }
      for (int j = eitr->os() - 1; j <= eitr->oe() - 1; ++j) fconv[j] = true;
      if (nii != 2) statvec.resize(nii);
      int cno = 1;
      for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++cno)
      {
        ditr->resize(id.ni());
        if (opt.cs() <= cno && cno <= opt.ce() && opt.smatch(*ditr)) cal_convi(eitr, statvec, vstatmap, *ditr);
      }
      if (etype == fncordev || etype == fncorline || etype == fncormid) set_statvec((eitr->carglst()).begin(), statvec, vstatmap);
      cno = 1;
      for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++cno)
      {
        if (opt.cs() <= cno && cno <= opt.ce() && opt.smatch(*ditr))
        {
          ldata_itr ditr1 = ditr;
          ++ditr1;
          if (ditr1 == datalst.end() || cno == opt.ce()) lastd = true;
          if (etype == fncorline || etype == fncormid) didx = get_qc(opt, datalst, qcdata0, qcdata1, nd, qccode, ditr, sidx, ididx, cno, didx);
          conv_case(eitr, statvec, vstatmap, firstd, lastd, *ditr0, *ditr, *ditr1, qcdata0, qcdata1, didx, nd);
          if (firstd) firstd = false;
          ditr0 = ditr;
        }
      }
    }
  }
}

void conviclass::psub(string& targstr, const lcarg& carglst) const
// 部分文字列を置換して返す
{
  lcarg::const_iterator aitr = carglst.begin();
  while (aitr != carglst.end())
  {
    int nrep = strregrep(targstr, aitr->str1(), aitr->str2(), aitr->global());
    if (nrep < 0) throw runtime_error("文字列置換指定/"+aitr->str1()+"/"+aitr->str2()+"/が間違ってるよ！");
    ++aitr;
  }
}

void conviclass::set_statvec(lcarg::iterator citr, vecstatv& statvec, map<string, vecstatv>& vstatmap) const
// 基準値をセットする
{
  if (citr->low() != BLANKf)    // 基準値指定
  { for (size_t i = 0; i < statvec.size(); ++i) statvec[i].sigma(citr->low()); }
  else
  {
    for (map<string, vecstatv>::iterator mitr = vstatmap.begin(); mitr != vstatmap.end(); ++mitr)
    {
      for (size_t i = 0; i < mitr->second.size(); ++i) statvec[i] += mitr->second[i];
    }
  }
}

int conviclass::get_qc(const option& opt, ldata& datalst, datac& qcdata0, datac& qcdata1, int& nd, string qccode, ldata_itr ditr, int sidx, int ididx, int cno, int didx) const
// QCデータを取得して現在のデータインデックスを返す
{
  string scode(ditr->sdata(sidx));
  if (strpmatch(ditr->sdata(ididx), qccode) > 0) // QCデータ
  {
    qcdata0 = *ditr;  qcdata1 = *ditr; // 直前と直後のQCデータ
    didx = 0;   nd = 0;
    int tcno = cno + 1;
    bool qc = false;
    ldata_itr tditr = ditr;
    ++tditr;
    while (qc == false && tditr != datalst.end())
    {
      if (opt.cs() <= tcno && tcno <= opt.ce() && opt.smatch(*tditr))
      {
        if (tditr->sdata(sidx) != scode)
        { qc = true; }
        else if (strpmatch(tditr->sdata(ididx), qccode) > 0)
        { qcdata1 = *tditr;  qc = true; }
        else
        { ++nd; }
      }
      ++tditr;  ++tcno;
    }
  }
  else
  { ++didx; }
  return didx;                  // 現在のデータインデックスを返す
}

void conviclass::cal_convi(lexpr::iterator eitr, vecstatv& statvec, map<string, vecstatv>& vstatmap, datac& cdata) const
// 項目データ変換用にデータを集計する
{
  lcarg::iterator citr = (eitr->carglst()).begin();
  if (eitr->type() == fncordev || eitr->type() == fncorline || eitr->type() == fncormid)
  { cal_cor(citr, statvec, vstatmap, cdata); }
  else if (eitr->type() == fncomp || eitr->type() == fncomps)
  { cal_comp(citr, statvec, cdata); }
  else if (eitr->type() == fnrand)
  {
    int osidx = eitr->os() - 1, oeidx = eitr->oe() - 1;
    int k = 0;
    for (int i = osidx; i <= oeidx; ++i, ++k)
    {
      double fd = (double)(rand());
      cdata.assign(i, fd);
      statvec[k].sigma(fd);
    }
  }
  else if (eitr->type() == fnrandn)
  {
    int osidx = eitr->os() - 1, oeidx = eitr->oe() - 1;
    int k = 0;
    for (int i = osidx; i <= oeidx; ++i, ++k)
    {
      double fd = (double)(randn(0.0, 1.0));
      cdata.assign(i, fd);
      statvec[k].sigma(fd);
    }
  }
  else                          // stat系関数
  {
    int isidx = citr->is() - 1, ieidx = citr->ie() - 1;
    int k = 0;
    for (int i = isidx; i <= ieidx; ++i, ++k)
    {
      if (cdata.fdata(i) != BLANKf) statvec[k].sigma(cdata.fdata(i));
    }
  }
}

void conviclass::cal_cor(lcarg::iterator citr, vecstatv& statvec, map<string, vecstatv>& vstatmap, datac& cdata) const
// cor.*()用にデータを集計する
{
  int sidx = citr->is() - 1;    // 層別項目のインデックス
  if (cdata.sdata(sidx) == BLANKs) return;
  ++citr;
  int ididx = citr->is() - 1; // ID項目のインデックス
  if (strpmatch(cdata.sdata(ididx), citr->str1()) < 1) return; // QCデータ以外
  ++citr;
  int isidx = citr->is() - 1, ieidx = citr->ie() - 1;
  int nii = ieidx - isidx + 1;
  map<string, vecstatv>::iterator mitr = vstatmap.find(cdata.sdata(sidx));
  if (mitr == vstatmap.end()) // 新規コード
  {
    vecstatv tvstat(nii);
    int k = 0;
    for (int i = isidx; i <= ieidx; ++i, ++k)
    {
      if (cdata.fdata(i) != BLANKf) tvstat[k].sigma(cdata.fdata(i));
    }
    vstatmap.insert(map<string, vecstatv>::value_type(cdata.sdata(sidx), tvstat));
  }
  else
  {
    int k = 0;
    for (int i = isidx; i <= ieidx; ++i, ++k)
    {
      if (cdata.fdata(i) != BLANKf) mitr->second[k].sigma(cdata.fdata(i));
    }
  }
}

void conviclass::cal_comp(lcarg::iterator citr, vecstatv& statvec, datac& cdata) const
// comp()用にデータを集計する
{
  double appr = cdata.fdata(citr->is() - 1);
  ++citr;
  double ab40r = cdata.fdata(citr->is() - 1);
  ++citr;
  double ab42 = 1.0;
  double eps = DBL_EPSILON * 5.0; // 計算誤差
  if (citr->type() == itemsub)
  {
    ab42 = cdata.fdata(citr->is() - 1);
    if (ab42 == BLANKf || fabs(ab42) < eps)
    { appr = ab40r = BLANKf; }
    else
    {
      if (appr != BLANKf) appr /= ab42;
      if (ab40r != BLANKf) ab40r /= ab42;
    }
  }
  if (appr != BLANKf) statvec[0].sigma(appr);
  if (ab40r != BLANKf) statvec[1].sigma(ab40r);
}

void conviclass::conv_case(lexpr::iterator eitr, vecstatv& statvec, map<string, vecstatv>& vstatmap, bool firstd, bool lastd, datac& cdata0, datac& cdata, datac& cdata1, datac& qcdata0, datac& qcdata1, int didx, int nd) const
// ケースデータを変換する
{
  lcarg::iterator citr = (eitr->carglst()).begin();
  int osidx = eitr->os() - 1, oeidx = eitr->oe() - 1;
  int isidx = citr->is() - 1, ieidx = citr->ie() - 1;
  int sidx = 0, ididx = 0, xidx = 0;
  if (eitr->type() == fndd)     // 中心差分変換のための変数xの項目インデックス
  {
    citr++;
    xidx = citr->is() - 1;
  }
  string qccode;
  cargtype catype = citr->type(); // 引数のタイプ
  map<string, vecstatv>::iterator mitr = vstatmap.begin();
  if (eitr->type() == fncordev || eitr->type() == fncorline || eitr->type() == fncormid)
  {
    sidx = citr->is() - 1;
    mitr = vstatmap.find(cdata.sdata(sidx));
    ++citr;
    ididx = citr->is() - 1;
    qccode = citr->str1();
    ++citr;
    isidx = citr->is() - 1;  ieidx = citr->ie() - 1;
  }
  int i = isidx, k = 0;
  int iinc = (isidx < ieidx) ? 1 : 0;
  for (int oidx = osidx; oidx <= oeidx; ++oidx)
  {
    switch (eitr->type())
    {
    case fncordev : conv_cordev(cdata, statvec, vstatmap, mitr, sidx, i, oidx, k, citr->ctype());
      break;
    case fncorline :
    case fncormid : conv_corlm(cdata, qcdata0, qcdata1, statvec, vstatmap, mitr, sidx, i, oidx, k, didx, nd, catype, citr->ctype(), eitr->type());
      break;
    case fncomp :
    case fncomps : conv_comp(cdata, eitr->type(), statvec, citr, isidx, oidx);
      break;
    case fnrand :
    case fnrandn : conv_rand(cdata, statvec, oidx, oidx - osidx);
      break;
    default : conv_stat(firstd, lastd, cdata0, cdata, cdata1, statvec, eitr, i, oidx, k, xidx);
    }
    i += iinc;  k += iinc;
  }
}

void conviclass::conv_cordev(datac& cdata, vecstatv& statvec, map<string, vecstatv>& vstatmap, map<string, vecstatv>::iterator mitr, int sidx, int idx, int oidx, int kidx, char cormeth) const
// ケースデータをcordev()変換する
{
  double fidx = cdata.fdata(idx);
  double eps = DBL_EPSILON * 5.0; // 計算誤差
  if (cdata.sdata(sidx) == BLANKs || mitr == vstatmap.end())
  { cdata.assign(oidx, cdata.sdata(idx)); }
  else if (fidx == BLANKf || mitr->second[kidx].n() < 1)
  { cdata.assign(oidx, cdata.sdata(idx)); }
  else
  {
    double qclevel = statvec[kidx].mean(); // 基準値
    if (cormeth == 'r')         // 率補正(比例尺度)
    {
      if (fabs(mitr->second[kidx].mean()) < eps) fidx = BLANKf;
      else fidx *= qclevel / mitr->second[kidx].mean();
    }
    else                        // 偏差補正(間隔尺度)
    { fidx -= mitr->second[kidx].mean() - qclevel; }
    cdata.assign(oidx, fidx);
  }
}

void conviclass::conv_corlm(datac& cdata, datac& qcdata0, datac& qcdata1, vecstatv& statvec, map<string, vecstatv>& vstatmap, map<string, vecstatv>::iterator mitr, int sidx, int idx, int oidx, int kidx, int didx, int nd, cargtype catype, char cormeth, fntype etype) const
// ケースデータをcorline()・cormid()変換する
{
  double fidx = cdata.fdata(idx);
  double eps = DBL_EPSILON * 5.0; // 計算誤差
  if (cdata.sdata(sidx) == BLANKs || mitr == vstatmap.end())
  { cdata.assign(oidx, cdata.sdata(idx)); }
  else if (fidx == BLANKf || mitr->second[kidx].n() < 1)
  { cdata.assign(oidx, cdata.sdata(idx)); }
  else if (qcdata0.fdata(idx) == BLANKf || qcdata1.fdata(idx) == BLANKf)
  { cdata.assign(oidx, cdata.sdata(idx)); }
  else
  {
    double qclevel = statvec[kidx].mean(); // 基準値指定またはtotal指定
    if (catype == itemsub) qclevel = mitr->second[kidx].mean(); // 層平均値指定
    if (cormeth == 'r')         // 率補正(比例尺度)
    {
      double qc0 = qcdata0.fdata(idx), qc1 = qcdata1.fdata(idx);
      double qci = (etype == fncorline) ? qc0 + (qc1 - qc0) / double(nd + 1) * didx : (qc0 + qc1) * 0.5;
      if (fabs(qci) < eps) fidx = BLANKf;
      else fidx *= qclevel / qci;
    }
    else
    {
      double dev0 = qcdata0.fdata(idx) - qclevel;
      double dev1 = qcdata1.fdata(idx) - qclevel;
      double devi = (etype == fncorline) ? dev0 + (dev1 - dev0) / double(nd + 1) * didx : (dev0 + dev1) * 0.5;
      fidx -= devi;
    }
    cdata.assign(oidx, fidx);
  }
}

void conviclass::conv_comp(datac& cdata, fntype ftype, vecstatv& statvec, lcarg::iterator citr, int isidx, int oidx) const
// ケースデータをcomp()変換する
{
  double appr = cdata.fdata(isidx);
  ++citr;
  double ab40r = cdata.fdata(citr->is() - 1);
  ++citr;
  double ab42 = 1.0;
  double eps = DBL_EPSILON * 5.0; // 計算誤差
  if (citr->type() == itemsub)
  {
    ab42 = cdata.fdata(citr->is() - 1);
    if (ab42 == BLANKf || fabs(ab42) < eps)
    { appr = ab40r = BLANKf; }
    else
    {
      if (appr != BLANKf) appr /= ab42;
      if (ab40r != BLANKf) ab40r /= ab42;
    }
  }
  double composit = BLANKf;
  if (appr != BLANKf && ab40r != BLANKf && fabs(statvec[0].sd()) >= eps && fabs(statvec[1].sd()) >= eps)
  {
    double mappr = (ftype == fncomps) ? Mappr : statvec[0].mean();
    double sdappr = (ftype == fncomps) ? SDappr : statvec[0].sd();
    double mab40r = (ftype == fncomps) ? Mab40r : statvec[1].mean();
    double sdab40r = (ftype == fncomps) ? SDab40r : statvec[1].sd();
    composit = 0.5 * ((appr - mappr) / sdappr + (ab40r - mab40r) / sdab40r);
  }
  cdata.assign(oidx, composit);
}

void conviclass::conv_rand(datac& cdata, vecstatv& statvec, int oidx, int kidx) const
// 乱数を出力する
{
  double fd = cdata.fdata(oidx);
  fd = (statvec[kidx].sd() > 0.0) ? (fd - statvec[kidx].mean()) / statvec[kidx].sd() : BLANKf;
  cdata.assign(oidx, fd);
}

void conviclass::conv_stat(bool firstd, bool lastd, datac& cdata0, datac& cdata, datac& cdata1, vecstatv& statvec, lexpr::iterator eitr, int idx, int oidx, int kidx, int xidx) const
// ケースデータをstat系関数変換する
{
  if (statvec[kidx].n() < 1)   // 基礎統計量なし
  { cdata.assign(oidx, BLANKs); }
  else
  {
    double fd = cdata.fdata(idx);
    if (fd != BLANKf)
    {
      switch (eitr->type())
      {
      case fncum :
        if (firstd == false) fd += cdata0.fdata(oidx);
        break;
      case fndev : fd -= statvec[kidx].mean();
        break;
      case fndif :
        if (firstd == false) fd -= cdata0.fdata(idx);
        break;
      case fndd :
        if (firstd || lastd) fd = BLANKf;
        else fd = cal_dd(cdata0, cdata, cdata1, idx, xidx);
        break;
      case fnmean : fd = statvec[kidx].mean();
        break;
      case fnnorm :
        fd = (statvec[kidx].sd() > 0.0) ? (fd - statvec[kidx].mean()) / statvec[kidx].sd() : BLANKf;
        break;
      case fnsd : fd = statvec[kidx].sd();
        break;
      }
    }
    cdata.assign(oidx, fd);
  }
}

double conviclass::cal_dd(datac& cdata0, datac& cdata, datac& cdata1, int idx, int xidx) const
// 中心差分変換値を返す
{
  double y0 = cdata0.fdata(idx), y = cdata.fdata(idx), y1 = cdata1.fdata(idx);
  double x0 = cdata0.fdata(xidx), x = cdata.fdata(xidx), x1 = cdata1.fdata(xidx);
  if (y0 == BLANKf || y == BLANKf || y1 == BLANKf || x0 == BLANKf || x == BLANKf || x1 == BLANKf) return BLANKf;
  double dx0 = x - x0, dx1 = x1 - x;
  double eps = DBL_EPSILON * fmax(5.0, fabs(dx0)); // 計算誤差
  double w0 = (fabs(dx0) < eps) ? 0.0 : 1.0 / dx0;
  double w1 = (fabs(dx1) < eps) ? 0.0 : 1.0 / dx1;
  double b0 = (y - y0) * w0, b1 = (y1 - y) * w1;
  w0 = fabs(w0);  w1 = fabs(w1);
  if (fabs(w0 + w1) < eps) return 0.0;
  return (b0 * w0 + b1 * w1) / (w0 + w1); // 比例配分した中心差分を返す
}

int conviclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込んで保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw option_error("実行時エラー！");
  int imax = opt.check(id.ni()); // オプションと変換後最大項目のチェック
  if (imax < 0) throw option_error("実行時オプションエラー！");
  if (imax > (int)(id.ni())) id.set(id.title(), imax); // 項目数を増やす
  return nc;
}

void conviclass::put(option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return;
  vecbool fconv(id.ni(), false);
  conv_data(opt, id, datalst, fconv);
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if (opt.prec() < 1)
    { ditr->put(cout); }
    else
    {
      int prec0 = cout.precision(); // 現在の浮動小数点精度を保存する
      cout.precision(opt.prec());   // 現在の浮動小数点精度を設定する
      if (fconv[0] == true && ditr->fdata(0) != BLANKf) cout << ditr->fdata(0);
      else cout << ditr->dstr(0);
      for (size_t j = 1; j < id.ni(); ++j)
      {
        if (fconv[j] == true && ditr->fdata(j) != BLANKf) cout << CHRdelm << ditr->fdata(j);
        else cout << CHRdelm << ditr->dstr(j);
      }
      cout.precision(prec0);
      cout << endl;
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  conviclass convi;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) convi.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { convi.get(argv[i], opt, id, datalst);  fok = false; }
    }
    convi.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
