// convs.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-11-24 15:30:26 tac>
//
// dansのデータファイルを入力し、文字列データの各種変換を行う。
//
// 使用ライブラリ:libfun.so, libdans.so
// コンパイル法 :g++ -O -s -W convs convs.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)
#include <boost/format.hpp>     // format用ヘッダファイル

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "convs  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:convs [\"-<変換式>\" …] [<オプション> …] [<データファイル> …]",
  "機能:指定された変換式に従って文字列データの変換をするんよ。",
  "変換式:i[<変換後データ出力開始項目番号>][-[<終了項目番号>]]=<文字列関数>",
  "識別情報関数:item(),title()",
  "文字列関数:anhan(),anzen(),cat(),date(),day(),dino(),div(),dum()",
  "  first(),gnum(),grade(),han(),kanahan(),kanazen(),last(),low()",
  "  max(),min(),nd(),num(),part(),psub(),rand(),real(),sub()",
  "  tfirst(),tlast(),trm(),up(),zen()",
  "<オプション>",
  "  -c=[<開始ケース番号>][-[<終了ケース番号>]]:変換ケース番号の指定。",
  "  -cd[={-|+}]:ケースデータの出力指定なんよ。(-cd:-cd=+)",
  "  -f=<ファイル名>:変換式やオプションを指定したファイルから読み込むんよ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんよ。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定なんよ。(-id:-id=+)",
  "  -r<層別項目番号>=[<実数下限>][-[<実数上限>]]:実数層別項目指定",
  "  -s<層別項目番号>=<文字列>[,<文字列>,…]:文字層別項目指定",
  "    指定した実数範囲内または文字列と一致するケースだけ変換するんよ。"
};

enum fntype                     // 関数の種類
{
  fnblank,                      // 指定なし
  fnanhan,                      // カタカナ以外半角変換関数
  fnanzen,                      // カタカナ以外全角変換関数
  fncat,                        // 書式指定変換関数
  fndate,                       // 日付変換関数
  fnday,                        // 日数変換関数
  fndino,                       // 項目番号変換関数
  fndiv,                        // 分割関数
  fndum,                        // ダミー項目変換関数
  fnfirst,                      // 欠損ではない最初のデータ取得関数
  fngnum,                       // 群内番号出力関数
  fngrade,                      // グレード変換関数
  fnhan,                        // 半角変換関数
  fnitem,                       // 項目名変換関数
  fnkanahan,                    // カタカナ半角変換関数
  fnkanazen,                    // カタカナ全角変換関数
  fnlast,                       // 欠損ではない最後のデータ取得関数
  fnlow,                        // 小文字変換関数
  fnmax,                        // 最大値変換関数
  fnmin,                        // 最小値変換関数
  fnnd,                         // データ数取得関数
  fnnum,                        // ケース番号出力関数
  fnpart,                       // 部分文字列取得関数
  fnpsub,                       // 部分文字列変換関数
  fnrand,                       // 乱数出力関数
  fnreal,                       // 実数変換関数
  fnsub,                        // 文字列変換関数
  fntfirst,                     // 欠損ではない最初のデータ番号取得関数
  fntlast,                      // 欠損ではない最後のデータ番号取得関数
  fntitle,                      // データ名変換関数
  fntrm,                        // トリミング関数
  fnup,                         // 大文字変換関数
  fnzen,                        // 全角変換関数
  fnetc                         // その他の関数
};

enum cargtype                   // 引数の種類
{
  cargblank,                    // 引数なし
  itemsub,                      // 項目
  itemapp,                      // 項目名追加
  sdata,                        // 文字列データ
  cargetc                       // その他
};

class carg                      // 引数クラス
{
private:
  cargtype vtype;               // 引数の種類
  pairint irange;               // 変換項目の開始・終了番号
  string vstr1;                 // 第1文字列(変換対象文字列等)
  string vstr2;                 // 第2文字列(出力指定・変換後文字列等)
  bool vglobal;                 // グローバル置換フラッグ
  char vctype;                  // 変換指定型('d','f','g','s','x'等)
  pairdbl frange;               // 実数の上・下限(範囲指定等)
public:
  carg() : vtype(cargblank), irange(0, 0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // デフォールトコンストラクタ
  {}
  explicit carg(cargtype type0, int is0 = 0, int ie0 = 0) : vtype(type0), irange(is0, ie0), vstr1(), vstr2(), vglobal(false), vctype(' '), frange(BLANKf, BLANKf)
  // コンストラクタ
  {}
  carg(cargtype type0, int is0, int ie0, const string& str10, const string& str20, bool global0 = false, char ctype0 = ' ', double low0 = BLANKf, double up0 = BLANKf) : vtype(type0), irange(is0, ie0), vstr1(str10), vstr2(str20), vglobal(global0), vctype(ctype0), frange(low0, up0)
  // コンストラクタ
  {}
  carg(const carg& carg0) : vtype(carg0.vtype), irange(carg0.irange), vstr1(carg0.vstr1), vstr2(carg0.vstr2), vglobal(carg0.vglobal), vctype(carg0.vctype), frange(carg0.frange)
  // コピーコンストラクタ
  {}
  virtual ~carg()               // デストラクタ
  {}
  const carg& operator=(const carg& carg0); // コピー代入演算子
  cargtype type() const                     // 引数の種類を返す
  { return vtype; }
  cargtype type(cargtype type0) // 引数の種類を設定する
  { return (vtype = type0); }
  int is() const                // 変換項目の開始番号を返す
  { return irange.first; }
  int is(int is0)               // 変換項目の開始番号を設定する
  { return (irange.first = is0); }
  int ie() const                // 変換項目の終了番号を返す
  { return irange.second; }
  int ie(int ie0)               // 変換項目の終了番号を設定する
  { return (irange.second = ie0); }
  const string& str1() const    // 文字列1を返す
  { return vstr1; }
  const string& str1(const string& str10) // 文字列1を設定する
  { return (vstr1 = str10); }
  const string& str2() const    // 文字列2を返す
  { return vstr2; }
  const string& str2(const string& str20) // 文字列2を設定する
  { return (vstr2 = str20); }
  bool global() const           // グローバル置換フラッグを返す
  { return vglobal; }
  bool global(bool global0)     // グローバル置換フラッグを設定する
  { return (vglobal = global0); }
  char ctype() const            // 変換指定型を返す
  { return vctype; }
  char ctype(char ctype0)       // 変換指定型を設定する
  { return (vctype = ctype0); }
  double low() const            // 実数の下限を返す
  { return frange.first; }
  double low(double low0)       // 実数の下限を設定する
  { return (frange.first = low0); }
  double up() const             // 実数の上限を返す
  { return frange.second; }
  double up(double up0)         // 実数の上限を設定する
  { return (frange.second = up0); }
};

const carg& carg::operator=(const carg& carg0) // コピー代入演算子
{
  if (&carg0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = carg0.vtype;  irange = carg0.irange;
  vstr1 = carg0.vstr1;  vstr2 = carg0.vstr2;  vglobal = carg0.vglobal;
  vctype = carg0.vctype;  frange = carg0.frange;
  return *this;
}

using lcarg = list<carg>;       // 引数リストの別名

class expr                      // 変換式クラス
{
private:
  fntype vtype;                 // 関数の種類
  pairint irange;               // 出力項目の開始・終了番号
  lcarg vcarglst;          // 引数リスト
  cargtype get_target(int& is, int& ie, string& targstr, const string& sstr) const;
  // 変換対象項目または変換対象文字列を取得する
  int get_expr(const string& pname, const string& sstr); // 変換式を取得する
  fntype search_fn(const string& str) const; // 関数をサーチする
  int get_cat(const string& pname, const string& cargstr);
  // cat()関数の引数を取得する(文字列と書式指定)
  int get_div(const string& pname, const string& cargstr);
  // div()関数の引数を取得する(項目分割指定)
  int get_dum(const string& pname, const string& cargstr);
  // dum()関数の引数を取得する(ダミー項目指定)
  int get_grade(const string& pname, const string& cargstr);
  // grade()関数の引数を取得する(グレード指定)
  int get_part(const string& pname, const string& cargstr);
  // part()関数の引数を取得する(部分文字列指定)
  int get_sub(const string& pname, const string& cargstr, bool subf);
  // sub()・psub()関数の引数を取得する(置換文字列指定)
  int get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr, bool subf) const;
  // 置換対象文字列と置換後文字列とグローバル指定を取得する
  int get_title(const string& pname, const string& cargstr);
  // title()関数の引数を取得する(置換文字列指定)
public:
  expr() : vtype(fnblank), irange(0, 0), vcarglst() // デフォールトコンストラクタ
  {}
  expr(fntype type0, int os0, int oe0, const lcarg& carglst0) : vtype(type0), irange(os0, oe0), vcarglst(carglst0)
  // コンストラクタ
  {}
  expr(const expr& expr0) : vtype(expr0.vtype), irange(expr0.irange), vcarglst(expr0.vcarglst)
  // コピーコンストラクタ
  {}
  virtual ~expr()               // デストラクタ
  {}
  const expr& operator=(const expr& expr0); // コピー代入演算子
  fntype type() const                       // 関数の種類を返す
  { return vtype; }
  fntype type(fntype type0)     // 関数の種類を設定する
  { return (vtype = type0); }
  int os() const                // 出力項目の開始番号を返す
  { return irange.first; }
  int os(int os0)               // 出力項目の開始番号を設定する
  { return (irange.first = os0); }
  int oe() const                // 出力項目の終了番号を返す
  { return irange.second; }
  int oe(int oe0)               // 出力項目の終了番号を設定する
  { return (irange.second = oe0); }
  lcarg& carglst()         // 引数リストを返す
  { return vcarglst; }
  int get(const string& pname, const string& oistr, const string& exstr);
  // 変換式を取り出す
};

const expr& expr::operator=(const expr& expr0) // コピー代入演算子
{
  if (&expr0 == this) return *this; // 自分自身への代入を防ぐ
  vtype = expr0.vtype;  irange = expr0.irange;  vcarglst = expr0.vcarglst;
  return *this;
}

int expr::get(const string& pname, const string& oistr, const string& exstr)
// 変換式を取得する
{
  if (oistr.size() > 0)
  {
    string tstr;
    if (get_target(irange.first, irange.second, tstr, oistr) == cargblank) return disperr(pname, "出力項目指定が間違っとるんよ→", oistr);
  }
  int err = get_expr(pname, exstr);
  if (err) return err;
  if (vtype == fntitle && oistr.size() > 0) return disperr(pname, "title()は単独で使用するんよ→", exstr);
  return 0;                     // 取得成功
}

cargtype expr::get_target(int& is, int& ie, string& targstr, const string& sstr) const
// 変換対象項目または変換対象文字列を取得する
{
  targstr = sstr;
  strtrm(targstr);
  if (targstr.size() < 1) return cargblank;
  is = 0;  ie = 0;
  if (targstr[0] == 'i')        // 項目指定の場合
  {
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(1)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemsub;
  }
  if (targstr.size() > 1 && targstr[0] == '+' && targstr[1] == 'i')
  {                             // 項目名追加の場合
    is = 1;  ie = INT_MAX;
    if (getrange(is, ie, targstr.substr(2)) < 1) return cargblank;
    if (is < 1 || is > ie) return cargblank;
    return itemapp;
  }
  sstrtostr(targstr);           // 特殊表現文字を変換し、""を削除する
  return sdata;                 // 文字列指定の場合
}

int expr::get_expr(const string& pname, const string& sstr) // 変換式を取得する
{
  string tstr(sstr), substr;
  strsize_t idx = getsubstr(substr, tstr, '(');
  if (idx == string::npos || idx >= tstr.size()) return disperr(pname, "左括弧[(]がないんよ→", sstr);
  strtrm(substr);
  strlower(substr);             // 関数名を小文字にする
  if ((vtype = search_fn(substr)) == fnblank) return disperr(pname, "こんな関数ないんよ→", sstr);
  tstr.erase(0, idx + 1);
  idx = getsubstr(substr, tstr, ')');
  if (idx == string::npos || idx >= tstr.size()) return disperr(pname, "右括弧[)]がないんよ→", sstr);
  strtrm(substr);
  int n = 0, is = 0, ie = 0;
  cargtype atype = cargblank;
  string targstr;
  carg tag;
  bool subf = true;
  switch (vtype)
  {
  case fncat : n = get_cat(pname, substr);
    break;
  case fndiv : n = get_div(pname, substr);
    if ((irange.first != irange.second) && (irange.second - irange.first + 1 != n))
    { return disperr(pname, "div()の出力項目は最初の項目番号だけ指定するんよ→", sstr); }
    if (n > 1) irange.second = irange.first + n - 1; // 出力項目は分割数分必要
    break;
  case fndum : n = get_dum(pname, substr);
    if ((irange.first != irange.second) && (irange.second - irange.first + 2 != n))
    { return disperr(pname, "dum()の出力項目は最初の項目番号だけ指定するんよ→", sstr); }
    if (n > 1) irange.second = irange.first + n - 2; // ダミー項目は(コード数ー1)必要
    break;
  case fngrade : n = get_grade(pname, substr);
    break;
  case fngnum :
  case fnfirst :
  case fnlast :
  case fnmax :
  case fnmin :
  case fnnd :
  case fntfirst :
  case fntlast : atype = get_target(is, ie, targstr, substr);
    if (atype != itemsub)
    { n = disperr(pname, "この関数の変換指定は項目だけなんよ→", sstr); }
    else
    {
      tag.type(atype);  tag.is(is);  tag.ie(ie);  tag.str1(targstr);
      vcarglst.push_back(tag);
    }
    break;
  case fnnum : n = 0;
    if (substr.size() > 0) n = disperr(pname, "この関数に引数はいらないんよ→", sstr);
    break;
  case fnrand : atype = get_target(is, ie, targstr, substr);
    if (atype != sdata)
    { n = disperr(pname, "変換指定項目が間違っとるんよ→", sstr); }
    else
    {
      char* prt = nullptr;
      int seed = strtol(targstr.c_str(), &prt, 0);
      srand(seed);              // 乱数の種をまく
    }
    break;
  case fnpart : n = get_part(pname, substr);
    break;
  case fnitem :
  case fnpsub :
    subf = false;
    n = get_sub(pname, substr, subf);
    break;
  case fnsub : n = get_sub(pname, substr, subf);
    break;
  case fntitle : n = get_title(pname, substr);
    break;
  default : atype = get_target(is, ie, targstr, substr); // その他の引数
    if (atype == cargblank)
    { n = disperr(pname, "変換指定項目が間違っとるんよ→", sstr); }
    else
    {
      tag.type(atype);  tag.is(is);  tag.ie(ie);  tag.str1(targstr);
      vcarglst.push_back(tag);
    }
  }
  if (n < 0) vtype = fnblank;
  return (n < 0) ? n : 0;       // 取得成功は0を返す
}

fntype expr::search_fn(const string& str) const // 関数をサーチする
{
  if (str.empty() || str == "cat") return fncat;
  if (str == "anhan") return fnanhan;
  if (str == "anzen") return fnanzen;
  if (str == "date") return fndate;
  if (str == "day") return fnday;
  if (str == "dino") return fndino;
  if (str == "div") return fndiv;
  if (str == "dum") return fndum;
  if (str == "first") return fnfirst;
  if (str == "gnum") return fngnum;
  if (str == "grade") return fngrade;
  if (str == "han") return fnhan;
  if (str == "item") return fnitem;
  if (str == "kanahan") return fnkanahan;
  if (str == "kanazen") return fnkanazen;
  if (str == "last") return fnlast;
  if (str == "low") return fnlow;
  if (str == "max") return fnmax;
  if (str == "min") return fnmin;
  if (str == "nd") return fnnd;
  if (str == "num") return fnnum;
  if (str == "part") return fnpart;
  if (str == "psub") return fnpsub;
  if (str == "rand") return fnrand;
  if (str == "real") return fnreal;
  if (str == "sub") return fnsub;
  if (str == "tfirst") return fntfirst;
  if (str == "tlast") return fntlast;
  if (str == "title") return fntitle;
  if (str == "trm") return fntrm;
  if (str == "up") return fnup;
  if (str == "zen") return fnzen;
  return fnblank;
}

int expr::get_cat(const string& pname, const string& cargstr)
// cat()関数の引数を取得する(文字列と書式指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) // cat()の場合
  {
    carg tag(sdata);            // 変換後文字列はヌル文字列になる
    vcarglst.push_back(tag);
    return 1;
  }
  int err = 0;
  lstr_itr itr = cargstrlst.begin();
  while (!err && itr != cargstrlst.end())
  {
    if (itr->empty())
    {
      err = disperr(pname, "cat()の変換対象指定が間違っとるんよ→", cargstr);
      break;
    }
    int is = 0, ie = 0, n = 0;
    string targstr, formstr;
    char ctype = 's';
    vecstr strvec;
    strsplit(strvec, *itr, CHRstr); // 変換対象指定と書式指定を分離
    cargtype atype = get_target(is, ie, targstr, strvec[0]);
    if (atype == cargblank || targstr.empty())
    { err = disperr(pname, "cat()の変換対象指定が間違っとるんよ→", cargstr); }
    else if (strvec.size() == 3) // 書式指定あり
    {
      formstr = strvec[1];
      if (formstr[0] != '%')
      { err = disperr(pname, "cat()の書式指定が間違っとるんよ→", cargstr); }
      else
      {
        if (formstr.find('d') != string::npos) ctype = 'd';
        else if (formstr.find('f') != string::npos) ctype = 'f';
        else if (formstr.find('g') != string::npos) ctype = 'g';
        else if (formstr.find('s') != string::npos) ctype = 's';
        else if (formstr.find('x') != string::npos || formstr.find('X') != string::npos) ctype = 'x';
        else err = disperr(pname, "cat()の書式指定が間違っとるんよ→", cargstr);
        char* ptr = nullptr;
        n = strtol((formstr.substr(1)).c_str(), &ptr, 0); // 出力幅取得
        if (n < 1) n = targstr.size();
      }
    }
    else if (strvec.size() != 1)
    { err = disperr(pname, "cat()の変換対象指定が間違っとるんよ→", cargstr); }
    if (!err)
    {
      carg tag(atype, is, ie, targstr, formstr, false, ctype, double(n), double(n));
      vcarglst.push_back(tag);
      ++n;
    }
    ++itr;
  }
  return err ? err : vcarglst.size(); // 成功の場合は引数リストの数を返す
}

int expr::get_div(const string& pname, const string& cargstr)
// div()関数の引数を取得する(項目分割指定)
{
  vecstr cargstrvec;
  if (strsplit(cargstrvec, cargstr, CHRstr) < 4) return disperr(pname, "div()の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string targstr;
  cargtype atype = get_target(is, ie, targstr, cargstrvec[0]);
  if (atype == cargblank || (atype == itemsub && is != ie)) return disperr(pname, "div()の変換対象指定が間違っとるんよ→", cargstr);
  string delmstr(cargstrvec[1]);
  sstrtostr(delmstr);           // 特殊表現文字を変換する
  char* ptr = nullptr;
  int ndiv = strtol(cargstrvec[2].c_str(), &ptr, 0);
  if (ndiv < 1) return disperr(pname, "div()の分割数指定が間違っとるんよ→", cargstr);
  carg tag(atype, is, ie, targstr, delmstr, false, ' ', double(ndiv), double(ndiv));
  vcarglst.push_back(tag);
  return ndiv;                  // 分割数を返す
}

int expr::get_dum(const string& pname, const string& cargstr)
// dum()関数の引数を取得する(ダミー項目指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 2) return disperr(pname, "dum()はコードが2つ以上必要なんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  vecstr cargstrvec;
  if (strsplit(cargstrvec, *itr, '=') < 2) return disperr(pname, "dum()の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string targstr;
  cargtype atype = get_target(is, ie, targstr, cargstrvec[0]);
  if (atype == cargblank || (atype == itemsub && is != ie)) return disperr(pname, "dum()の変換対象指定が間違っとるんよ→", cargstr);
  string codestr(cargstrvec[1]);
  sstrtostr(codestr);
  if (codestr[0] != CHRhead) codestr = CHRhead + codestr; // 完全一致にする
  if (codestr[codestr.size() -1] != CHRtail) codestr += CHRtail;
  carg tag(atype, is, ie, targstr, codestr, false, ' ', 0.0, 0.0);
  vcarglst.push_back(tag);
  ++itr;
  int n = 1;
  while (itr != cargstrlst.end())
  {
    codestr = *itr;
    sstrtostr(codestr);
    if (codestr[0] != CHRhead) codestr = CHRhead + codestr;
    if (codestr[codestr.size() -1] != CHRtail) codestr += CHRtail;
    tag.str2(codestr);  tag.low(double(n));  tag.up(double(n));
    vcarglst.push_back(tag);
    ++itr;  ++n;
  }
  return n;                     // コード数を返す
}

int expr::get_grade(const string& pname, const string& cargstr)
// grade()関数の引数を取得する(グレード指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "grade()の変換指定が間違っとるんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  vecstr cargstrvec;
  if (strsplit(cargstrvec, *itr, CHRstr) < 4) return disperr(pname, "grade()の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string targstr;
  cargtype atype = get_target(is, ie, targstr, cargstrvec[0]);
  if (atype == cargblank) return disperr(pname, "grade()の変換対象指定が間違っとるんよ→", cargstr);
  double low = -FLT_MAX, up = FLT_MAX;
  if (getrange(low, up, cargstrvec[1]) == 0 || low > up) return disperr(pname, "grade()のグレード範囲が変なんよ→", cargstr);
  string codestr(cargstrvec[2]);
  sstrtostr(codestr);
  carg tag(atype, is, ie, targstr, codestr, false, ' ', low, up);
  vcarglst.push_back(tag);
  ++itr;
  int err = 0, n = 1;
  while (!err && itr != cargstrlst.end())
  {
    cargstrvec.clear();
    if (strsplit(cargstrvec, *itr, CHRstr) < 4)
    { err = disperr(pname, "grade()の変換指定が間違っとるんよ→", cargstr); }
    else
    {
      low = -DBL_MAX;  up = DBL_MAX;
      if (getrange(low, up, cargstrvec[1]) == 0 || low > up)
      { err = disperr(pname, "grade()のグレード範囲が変なんよ→", cargstr); }
      else
      {
        codestr = cargstrvec[2];
        sstrtostr(codestr);
        tag.str2(codestr);  tag.low(low);  tag.up(up);
        vcarglst.push_back(tag);
        ++itr;  ++n;
      }
    }
  }
  return err ? err : n;         // エラーでなければグレード数を返す
}

int expr::get_part(const string& pname, const string& cargstr)
// part()関数の引数を取得する(部分文字列指定)
{
  vecstr cargstrvec;
  if (strsplit(cargstrvec, cargstr, CHRstr) < 3) return disperr(pname, "part()の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string targstr;
  cargtype atype = get_target(is, ie, targstr, cargstrvec[0]);
  if (atype == cargblank) return disperr(pname, "part()の変換対象指定が間違っとるんよ→", cargstr);
  int low = 1, up = INT_MAX;
  if (getrange(low, up, cargstrvec[1]) == 0 || low > up) return disperr(pname, "part()の範囲指定が間違っとるんよ→", cargstr);
  carg tag(atype, is, ie, targstr, cargstrvec[1], false, ' ', double(low), double(up));
  vcarglst.push_back(tag);
  return 1;
}

int expr::get_sub(const string& pname, const string& cargstr, bool subf)
// sub()・psub()・item()関数の引数を取得する(置換文字列指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "変換内容が間違っとるんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  string targstr, pattern, repstr;
  bool global = false;
  if (get_pattern(targstr, pattern, repstr, global, *itr, subf) < 0) return disperr(pname, "最初の変換指定が間違っとるんよ→", cargstr);
  int is = 0, ie = 0;
  string tstr;
  cargtype atype = get_target(is, ie, tstr, targstr);
  if (atype != itemsub && atype != itemapp) return disperr(pname, "この関数の変換対象指定は項目だけなんよ→", cargstr);
  carg tag(atype, is, ie, pattern, repstr, global);
  vcarglst.push_back(tag);
  ++itr;
  int err = 0, n = 1;
  while (!err && itr != cargstrlst.end())
  {
    if (get_pattern(targstr, pattern, repstr, global, *itr, subf) < 0)
    { err = disperr(pname, "2番目以後の変換指定が間違っとるんよ→", cargstr); }
    else
    {
      tag.str1(pattern);  tag.str2(repstr);  tag.global(global);
      vcarglst.push_back(tag);
      ++itr;
    }
  }
  return err ? err : n;         // エラーでなければ置換数を返す
}

int expr::get_pattern(string& targstr, string& pattern, string& repstr, bool& global, const string& cargstr, bool subf) const
// 置換対象文字列と置換後文字列とグローバル指定を取得する
{
  vecstr cargstrvec;
  if (strsplit(cargstrvec, cargstr, CHRstr) != 4) return -1;
  targstr = cargstrvec[0];  pattern = cargstrvec[1];  repstr = cargstrvec[2];
  sstrtostr(pattern);  sstrtostr(repstr); // 特殊表現文字列を変換し""を削除する
  global = (cargstrvec[3] == "g") ? true : false;
  if (subf)                     // sub()関数の時(完全一致)
  {
    if (global) return -1;      // sub()関数ではgは指定できない
    if (pattern.empty())        // ヌル文字列の時は「^$」にする
    {
      pattern = CHRhead;
      pattern += CHRtail;
    }
    else                        // 前後に「^」と「$」を追加する
    {
      if (pattern[0] != CHRhead) pattern = CHRhead + pattern;
      if (pattern[pattern.size() -1] != CHRtail) pattern += CHRtail;
    }
  }
  return 0;
}

int expr::get_title(const string& pname, const string& cargstr)
// title()関数の引数を取得する(置換文字列指定)
{
  lstr cargstrlst;
  if (strsplit(cargstrlst, cargstr, CHRdelm) < 1) return disperr(pname, "変換指定が間違っとるんよ!→", cargstr);
  lstr_itr itr = cargstrlst.begin();
  string targstr, pattern, repstr;
  bool global = false;
  int err = 0, n = 0;
  while (!err && itr != cargstrlst.end())
  {
    if (get_pattern(targstr, pattern, repstr, global, *itr, false) < 0)
    { err = disperr(pname, "変換指定が間違っとるんよ→", cargstr); }
    else
    {
      carg tag(sdata, 0, 0, pattern, repstr, global);
      vcarglst.push_back(tag);
      ++itr;
    }
  }
  return err ? err : n;         // エラーでなければ置換数を返す
}

using lexpr = list<expr>;       // 変換式リストの短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  pairint crange;               // 変換ケース番号
  bool vcd;                     // ケースデータ出力フラッグ
  bool vid;                     // データ識別情報出力フラッグ
  list<expr> vexlst;            // 変換式リスト
  lstra vstralst;               // 層別情報リスト
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int cs() const                // 変換開始ケース番号を返す
  { return crange.first; }
  int ce() const                // 変換終了ケース番号を返す
  { return crange.second; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  list<expr>& exlst()           // 変換式リストを返す
  { return vexlst; }
  lstra& stralst()              // 層別情報リストを返す
  { return vstralst; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);           // オプションのチェックとセット
  bool smatch(datac& cdata) const // 層別のチェック
  { return stramatch(vstralst, cdata); }
};

option::option(const char* c_pname) : vpath(), vpname(), crange(1, INT_MAX), vcd(true), vexlst(), vid(true), vstralst()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 変換ケース番号指定
  {
    if (getrange(crange.first, crange.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "cd")      // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0][0] == 'i')    // 変換式の指定
  {
    if (opt[1].empty())
    { err = badopt(vpname, optstr, USAGE); }
    else
    {
      expr tex;
      err = tex.get(vpname, opt[0], opt[1]);
      if (!err) vexlst.push_back(tex);
    }
  }
  else if (opt[0][0] == 'r' || opt[0][0] == 's') // 層別指定
  {
    stra tstra;
    char* prt = nullptr;
    int ino = strtol(opt[0].c_str() + 1, &prt, 0);
    if (tstra.get(opt[0][0], ino, opt[1]) == datanon) err = badopt(vpname, optstr, USAGE);
    if (!err) vstralst.push_back(tstra);
  }
  else if ((opt[0]).substr(0, 5) == "title") // データ名変更
  {
    if (opt[1].empty())
    {
      expr tex;
      err = tex.get(vpname, opt[1], opt[0]);
      if (!err) vexlst.push_back(tex);
    }
    else
    { err = badopt(vpname, optstr, USAGE); }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if (straok(vstralst, ni) == false) return disperr(vpname, "層別項目指定が間違っとるんよ！");
  bool err = false;
  int imax = ni;
  list<expr>::iterator itr = vexlst.begin();
  while (!err && itr != vexlst.end()) // 変換後項目番号最大値をサーチする
  {
    if (itr->oe() == INT_MAX) itr->oe(ni);
    else if (itr->oe() > imax) imax = itr->oe();
    if (itr->os() > itr->oe()) err = true;
    else ++itr;
  }
  if (err) return disperr(vpname, "出力項目指定が間違っとるんよ！");
  itr = vexlst.begin();
  while (!err && itr != vexlst.end())
  {
    lcarg::iterator aitr = (itr->carglst()).begin();
    while (!err && aitr != (itr->carglst()).end())
    {
      if (aitr->ie() == INT_MAX) aitr->ie(ni);
      if (aitr->is() > aitr->ie() || aitr->ie() > imax)
      { err = true; }
      else if ((itr->os() < itr->oe()) && (aitr->is() < aitr->ie()))
      {                         // 出力指定項目が複数の時
        if (itr->oe() - itr->os() != aitr->ie() - aitr->is()) err = true;
        else ++aitr;
      }
      else
      { ++aitr; }
    }
    if (!err) ++itr;
  }
  if (err) return disperr(vpname, "出力項目指定か変換元項目指定が間違っとるんよ！");
  return imax;                  // 変換後最大項目番号を返す
}

class convsclass                // convsメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, option& opt, idd& id) const;
  // データ識別情報の変換と出力
  void psub(string& targstr, const lcarg& carglst, bool subf) const;
  // 部分文字列を置換して返す
  void put_convscase(ostream& fout, option& opt, const idd& id, datac& cdata) const;
  // ケースデータを変換して出力する
  void cat(string& catstr, const lcarg& carglst, const datac& cdata, int offset) const;
  // 文字列を結合して返す
  void set_div(datac& tdata, const lcarg& carglst, int sidx, int eidx, const datac& cdata) const;
  // 項目を分割する
  void set_dum(datac& tdata, const lcarg& carglst, int sidx, int eidx, const datac& cdata) const;
  // ダミーデータに変換する
  lcarg::const_iterator search_targstr(const lcarg& carglst, const string& targstr) const;
  // 特殊文字を含む対象文字列のサーチ
  int fldata(string& tstr, fntype type, const lcarg& carglst, datac& cdata) const;
  // 欠損ではない最初/最後のデータ/データ番号を返す
  double stat(fntype type, const lcarg& carglst, datac& cdata) const;
  // 統計関数(最大値、最小値、データ数を返す)
  void convs_data(string& targstr, fntype type, const lcarg& carglst, const datac& cdata) const;
  // 文字列データを変換して返す
  int grade(string& targstr, const lcarg& carglst) const;
  // 数字データをグレード変換して返す
  double real(const string& targstr) const;
  // 文字列を可能な限り数字データに変換して返す
  int part(string& targstr, int sno, int eno) const;
  // 部分文字列を返す(漢字・半角カタカナを1文字と数える)
public:
  convsclass() : fname()        // コンストラクタ
  {}
  ~convsclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込んで保存する
  void put(option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void convsclass::put_id(ostream& fout, option& opt, idd& id) const
// データ識別情報の変換と出力
{
  string targstr;
  for (lexpr::iterator exitr = (opt.exlst()).begin(); exitr != (opt.exlst().end()); ++exitr)
  {
    if (exitr->type() == fntitle) // データ名変換
    {
      targstr = id.title();
      psub(targstr, exitr->carglst(), false);
      id.set(targstr, id.ni());
    }
    else if (exitr->type() == fnitem) // 項目名変換
    {
      int ino = 0;
      lcarg::const_iterator aitr = (exitr->carglst()).begin();
      if (aitr->type() == sdata) targstr = aitr->str1();
      else if (aitr->is() == aitr->ie()) targstr = id.item(aitr->is() - 1);
      else ino = aitr->is();
      for (int i = exitr->os(); i <= exitr->oe(); ++i)
      {
        if (ino > 0)            // 変換対象が複数項目の場合
        {
          targstr = id.item(ino - 1);
          ++ino;
        }
        string tstr(targstr);
        psub(tstr, exitr->carglst(), false);
        if (aitr->type() == itemapp) tstr = id.item(i - 1) + tstr;
        id.assign_item(i - 1, tstr);
      }
    }
  }
  id.put(fout, formtag);
}

void convsclass::psub(string& targstr, const lcarg& carglst, bool subf) const
// 部分文字列を置換して返す
{
  lcarg::const_iterator aitr = carglst.begin();
  while (aitr != carglst.end())
  {
    int nrep = strregrep(targstr, aitr->str1(), aitr->str2(), aitr->global());
    if (nrep < 0) throw runtime_error("文字列置換指定/" + aitr->str1() + "/" + aitr->str2() + "/が間違ってるよ！");
    if (subf && nrep > 0) aitr = carglst.end(); // sub()関数は最初にヒットしたところで終了する
    else ++aitr;
  }
}

void convsclass::put_convscase(ostream& fout, option& opt, const idd& id, datac& cdata) const
// ケースデータを変換して出力する
{
  static int ncc = 0;           // 変換データだけの通し番号
  static int ngc = 0;           // 変換データだけの群内通し番号
  static string gcode;          // 群コード
  datac tdata(cdata);
  for (lexpr::iterator exitr = (opt.exlst()).begin(); exitr != (opt.exlst()).end(); ++exitr)
  {
    int sidx = exitr->os() - 1, eidx = exitr->oe() - 1;
    lcarg::const_iterator aitr = (exitr->carglst()).begin();
    int idx = -1, j = 0;
    string targstr, tstr;
    switch (exitr->type())
    {
    case fntitle :
    case fnitem : break;        // title()とitem()は無視
    case fncat :
      for (j = sidx; j <= eidx; ++j)
      {
        cat(tstr, exitr->carglst(), cdata, j - sidx);
        tdata.assign(j, tstr);
      }
      break;
    case fndiv : set_div(tdata, exitr->carglst(), sidx, eidx, cdata);
      break;
    case fndum : set_dum(tdata, exitr->carglst(), sidx, eidx, cdata);
      break;
    case fngnum : targstr = cdata.sdata(aitr->is() - 1);
      if (ngc == 0 || (gcode != targstr)) { ngc = 1;  gcode = targstr; }
      else { ++ngc; }
      for (j = sidx; j <= eidx; ++j) tdata.assign(j, (double)ngc);
      break;
    case fnfirst :
    case fnlast :
    case fntfirst :
    case fntlast :
      for (j = sidx; j <= eidx; ++j)
      {
        idx = fldata(tstr, exitr->type(), exitr->carglst(), cdata);
        if (idx < 0) tdata.assign(j, BLANKs);
        else if (exitr->type() == fntfirst || exitr->type() == fntlast) tdata.assign(j, (double)idx);
        else tdata.assign(j, tstr);
      }
      break;
    case fnmax :
    case fnmin :
    case fnnd :
      for (j = sidx; j <= eidx; ++j)
      {
        double fdata = stat(exitr->type(), exitr->carglst(), cdata);
        if (fdata == BLANKf) tdata.assign(j, BLANKs);
        else tdata.assign(j, fdata);
      }
      break;
    case fnnum : ++ncc;
      for (j = sidx; j <= eidx; ++j) tdata.assign(j, (double)ncc);
      break;
    case fnrand : for (j = sidx; j <= eidx; ++j) tdata.assign(j, (double)(rand()));
      break;
    default : idx = -1;
      if (aitr->type() == sdata) targstr = aitr->str1();
      else if (aitr->is() == aitr->ie()) targstr = cdata.sdata(aitr->is() - 1);
      else idx = aitr->is() - 1;
      for (j = sidx; j <= eidx; ++j)
      {
        if (idx >= 0)
        {
          targstr = cdata.sdata(idx);
          ++idx;
        }
        tstr = targstr;
        convs_data(tstr, exitr->type(), exitr->carglst(), tdata);
        tdata.assign(j, tstr);
      }
    }
    for (j = sidx; j <= eidx; ++j) cdata.assign(j, tdata.sdata(j));
  }
  cdata.put(fout);
}

void convsclass::cat(string& catstr, const lcarg& carglst, const datac& cdata, int dist) const
// 文字列を結合して返す
{
  catstr.erase();
  for (auto cargi : carglst)
  {
    string tstr;
    if (cargi.type() == sdata)
    { tstr = cargi.str1(); }
    else
    {
      int idx = ((cargi.is() == cargi.ie()) ? cargi.is() : cargi.is() + dist) - 1;
      tstr = cdata.sdata(idx);
    }
    if (cargi.str2().size() > 0) // 変換指定あり
    {
      ostringstream ostr;
      int idata = 0;
      double fdata = BLANKf;
      char* ptr = nullptr;
      switch (cargi.ctype())
      {
      case 'd' :                // 整数変換
      case 'x' : idata = strtol(tstr.c_str(), &ptr, 0);
        if (ptr == tstr.c_str()) ostr << tstr;
        else ostr << boost::format((cargi.str2()).c_str()) % idata;
        break;
      case 'f' :                // 実数変換
      case 'g' : fdata = strtodf(tstr);
        if (fdata == BLANKf) ostr << tstr;
        else ostr << boost::format((cargi.str2()).c_str()) % fdata;
        break;
      default : ostr << boost::format((cargi.str2()).c_str()) % tstr;
      }
      tstr = ostr.str();
    }
    catstr += tstr;
  }
}

void convsclass::set_div(datac& tdata, const lcarg& carglst, int sidx, int eidx, const datac& cdata) const
// 項目を分割する
{
  lcarg::const_iterator aitr = carglst.begin();
  string tstr;
  if (aitr->type() == sdata) tstr = aitr->str1();
  else tstr = cdata.sdata(aitr->is() - 1);
  datac ddata(eidx - sidx + 1);
  ddata.getdata(tstr, (aitr->str2())[0]);
  for (int j = sidx; j <= eidx; ++j) tdata.assign(j, ddata.sdata(j - sidx));
}

void convsclass::set_dum(datac& tdata, const lcarg& carglst, int sidx, int eidx, const datac& cdata) const
// ダミーデータに変換する
{
  lcarg::const_iterator aitr = carglst.begin();
  string targstr;
  if (aitr->type() == sdata) targstr = aitr->str1();
  else targstr = cdata.sdata(aitr->is() - 1);
  if ((aitr = search_targstr(carglst, targstr)) == carglst.end())
  { for (int j = sidx; j <= eidx; ++j) tdata.assign(j, BLANKs); }
  else
  {
    for (int j = sidx; j <= eidx; ++j) tdata.assign(j, "0");
    if (aitr->low() > 0.0) tdata.assign(sidx + int(aitr->low()) - 1, "1");
  }
}

lcarg::const_iterator convsclass::search_targstr(const lcarg& carglst, const string& targstr) const
// 特殊文字を含む対象文字列のサーチ
{
  bool hit = false;
  lcarg::const_iterator aitr = carglst.begin();
  while (!hit && aitr != carglst.end())
  {
    if (strpmatch(targstr, aitr->str2()) != 0) hit = true;
    else ++aitr;
  }
  return aitr;                  // ヒットした要素の反復子を返す
}

int convsclass::fldata(string& tstr, fntype type, const lcarg& carglst, datac& cdata) const
// 欠損ではない最初/最後のデータ/データ番号を返す
{
  lcarg::const_iterator aitr = carglst.begin();
  int idx = -1, sidx = aitr->is() - 1, eidx = aitr->ie() - 1;
  tstr = BLANKs;
  for (int i = sidx; i <= eidx; ++i)
  {
    if (cdata.sdata(i) != BLANKs)
    {
      tstr = cdata.sdata(i);  idx = i - sidx;
      if (type == fnfirst || type == fntfirst) i = eidx + 1;
    }
  }
  return idx;                   // データがない時は-1を返す
}

double convsclass::stat(fntype type, const lcarg& carglst, datac& cdata) const
// 統計関数(最大値、最小値、データ数を返す)
{
  lcarg::const_iterator aitr = carglst.begin();
  double fmin = DBL_MAX, fmax = -DBL_MAX;
  int nd = 0, sidx = aitr->is() - 1, eidx = aitr->ie() - 1;
  for (int i = sidx; i <= eidx; ++i)
  {
    if (cdata.sdata(i) != BLANKs)
    {
      ++nd;                     // データ数は文字データも数える
      double fdata = cdata.fdata(i);
      if (fdata != BLANKf)
      {
        if (fdata < fmin) fmin = fdata;
        if (fdata > fmax) fmax = fdata;
      }
    }
  }
  switch (type)
  {
  case fnmax : return (fmax == -DBL_MAX) ? BLANKf : fmax;
  case fnmin : return (fmin == DBL_MAX) ? BLANKf : fmin;
  case fnnd : return (double)nd;
  }
  return BLANKf;
}

void convsclass::convs_data(string& targstr, fntype type, const lcarg& carglst, const datac& cdata) const
// 文字列データを変換して返す
{
  long d = 0;
  char* ptr = nullptr;
  double fdata = 0.0;
  ostringstream ostr;
  lcarg::const_iterator aitr = carglst.begin();
  switch (type)
  {
  case fnanhan : strzentohan(targstr, 'n'); // カタカナ以外を半角に変換する
    break;
  case fnanzen : strhantozen(targstr, 'n'); // カタカナ以外を全角に変換する
    break;
  case fndate : d = strtol(targstr.c_str(), &ptr, 0);
    if (d > 0) date(targstr, d);
    break;
  case fnday : d = day(targstr);
    if (d > 0)
    {
      ostr << d;
      targstr = ostr.str();
    }
    break;
  case fndino : d = strtol(targstr.c_str(), &ptr, 0);
    targstr = (d > 0) ? cdata.sdata(int(d - 1)) : BLANKs;
    break;
  case fngrade : grade(targstr, carglst);
    break;
  case fnhan : strzentohan(targstr, 'a'); // 半角に変換する
    break;
  case fnkanahan : strzentohan(targstr, 'k'); // カタカナだけを半角に変換する
    break;
  case fnkanazen : strhantozen(targstr, 'k'); // カタカナだけを全角に変換する
    break;
  case fnlow : strlower(targstr);
    break;
  case fnpart : part(targstr, int(aitr->low()), int(aitr->up()));
    break;
  case fnpsub : psub(targstr, carglst, false);
    break;
  case fnreal : fdata = real(targstr);
    if (fdata != BLANKf)
    {
      ostr << fdata;
      targstr = ostr.str();
    }
    break;
  case fnsub : psub(targstr, carglst, true);
    break;
  case fnup : strupper(targstr);
    break;
  case fntrm : strtrm(targstr);
    break;
  case fnzen : strhantozen(targstr, 'a'); // 全角に変換する
  }
}

int convsclass::grade(string& targstr, const lcarg& carglst) const
// 数字データをグレード変換して返す
{
  double fdata = strtodf(targstr);
  if (fdata == BLANKf) return -1; // 変換できない場合
  bool hit = false;
  lcarg::const_iterator aitr = carglst.begin();
  while (!hit && aitr != carglst.end())
  {
    if (aitr->low() <= fdata && fdata <= aitr->up()) hit = true;
    else ++aitr;
  }
  if (!hit) return 0;           // 変換せず
  targstr = aitr->str2();
  return 1;                     // 変換
}

int convsclass::part(string& targstr, int sno, int eno) const
// 部分文字列を返す(漢字・半角カタカナを1文字と数える)
{
  if (targstr.empty()) return 0; // ヌル文字列
  string tstr, buf;
  int nc = 0, no = 0, n = 0;
  strsize_t i = 0;
  while (no <= eno && i < targstr.size())
  {
    uchar c = targstr[i];
    if (nc > 1)                 // マルチバイトコードの途中
    { buf += c;  --nc; }
    else if (nc == 1)           // マルチバイトコードの最後
    {
      buf += c;  ++no;
      if (sno <= no && no <= eno) { tstr += buf;  ++n; }
      buf.erase();  nc = 0;
    }
    else if (ismcode(c))        // マルチバイトコード1バイト目
    { buf = c;  nc = ncode(c) - 1; }
    else
    {
      ++no;
      if (sno <= no && no <= eno) { tstr += c; ++n; }
    }
    ++i;
  }
  targstr = tstr;
  return n;                     // 取得した文字数を返す
}

double convsclass::real(const string& targstr) const
// 文字列を可能な限り数字データに変換して返す
{
  string tstr(targstr);
  strzentohan(tstr, 'a');       // まず全角を半角にする
  double fdata = BLANKf;
  sscanf(tstr.c_str(), "%lf", &fdata); // できるだけ数字に変換する
  return fdata;
}

int convsclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込んで保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw option_error("実行時エラー！");
  int imax = opt.check(id.ni()); // オプションと変換後最大項目のチェック
  if (imax < 0) throw option_error("実行時オプションエラー！");
  if (imax > (int)(id.ni())) id.set(id.title(), imax); // 項目数を増やす
  return nc;
}

void convsclass::put(option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return;
  int cno = 1;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr, ++cno)
  {
    ditr->resize(id.ni());
    if (opt.cs() <= cno && cno <= opt.ce() && opt.smatch(*ditr)) put_convscase(cout, opt, id, *ditr);
    else ditr->put(cout);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  convsclass convs;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) convs.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { convs.get(argv[i], opt, id, datalst);  fok = false; }
    }
    convs.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
