// sortd.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-11-05 16:39:11 tac>
//
// dansデータファイルのケースデータ部分をソートする。
// ソートのキー項目は、文字データとしても数字データとしてもソートできる。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法:g++ -O -s -W sortd sortd.C -I/usr/local/include -L/usr/local/lib
//             -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "sortd  Version 7.41          Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:sortd  [<オプション> …]  [<データファイル> …]",
  "機能:データファイルのケースデータ部分をソートします。",
  "<オプション>",
  "  -cd[={-|+}]:ケースデータの出力指定です。(-cd:-cd=+)",
  "  -key=[{r|s}]<第1キー項目番号>[,[{r|s}]<第2キー項目番号>,…]",
  "    キー項目の番号を指定します。このオプションは複数指定可能です。",
  "    r:数字データとしてソート(省略可)  s:文字データとしてソート",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定です。(-id:-id=+)",
  "  -sort={r|s}:ソート順を指定します。",
  "    r:逆順(大きい順)  s:正順(小さい順、オプション省略時)",
  "  -uniq[={-|f|l}]:キー項目が一致する重複ケースを削除します。(-uniq:-uniq=f)"
};

using pairkey = pair<datatype, int>; // キー項目の別名定義
using lpairkey = list<pairkey>;      // キー項目リストの別名定義

enum uniqtype                   // uniqオプションの種類
{
  uniqnon,                      // 指定なし
  uniqfirst,                    // 最初のケースを残す
  uniqlast,                     // 最後のケースを残す
  uniqetc                       // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vcd;                     // ケースデータ出力フラッグ
  lpairkey vkeylst;             // キー項目リスト
  bool vid;                     // データ識別情報出力フラッグ
  bool vsort;                   // ソート順フラッグ
  uniqtype vuniq;               // 重複ケース削除フラッグ
  int get_keylst(const string& sstr);
  // 文字列からキー項目指定を読み込みリストに追加する
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  const lpairkey& keylst() const // キー項目指定リストを返す
  { return vkeylst; }
  bool id() const               // データ識別情報力フラッグを返す
  { return vid; }
  bool sort() const             // ソート順フラッグを返す
  { return vsort; }
  uniqtype uniq() const         // 重複ケース削除フラッグを返す
  { return vuniq; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni) const;  // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vkeylst(), vid(true), vsort(true), vuniq(uniqnon)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "key")     // キー項目の指定
  {
    if (opt[1].empty()) vkeylst.clear(); // キー項目リストをクリアする
    else if (get_keylst(opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "sort")    // ソート順指定
  {
    if (opt[1].empty() || opt[1] == "s" ) vsort = true;
    else if (opt[1] == "r") vsort = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "uniq")    // 重複ケース削除指定
  {
    if (opt[1].empty() || opt[1] == "f") vuniq = uniqfirst;
    else if (opt[1] == "-") vuniq = uniqnon;
    else if (opt[1] == "f") vuniq = uniqfirst;
    else if (opt[1] == "l") vuniq = uniqlast;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::get_keylst(const string& sstr)
// 文字列からキー項目指定を読み込みリストに追加する
{
  lstr kstrlst;
  if (strsplit(kstrlst, sstr, CHRdelm) < 1) return 0; // 有効な範囲指定なし
  int size0 = vkeylst.size();
  for (string kstri : kstrlst)
  {
    pairkey key(dfdata, 0);
    if ((kstri)[0] == 'r') { key.first = dfdata;  kstri.erase(0, 1); }
    else if (kstri[0] == 's') { key.first = dsdata;  kstri.erase(0, 1); }
    else key.first = dfdata;     // 識別文字(r|s)なし
    char* ptr = nullptr;
    key.second = strtol(kstri.c_str(), &ptr, 0);
    if (key.second < 1 || ptr == kstri.c_str()) return -1;
    vkeylst.push_back(key);
  }
  return vkeylst.size() - size0; // 追加したキー項目指定数を返す
}

int option::check(size_t ni) const // オプションのチェックとセット
{
  if (vkeylst.empty()) return 0; // キー項目指定なし
  int err = 0;
  lpairkey::const_iterator kpos = vkeylst.begin();
  while (err == 0 && kpos != vkeylst.end())
  {
    if (kpos->second <= (int)ni) ++kpos;
    else err = disperr(vpname, "キー項目指定不適！");
  }
  return err;
}

class data_equal                // ケースデータの一致基準用関数オブジェクト
{
private:
  const lpairkey *keylst; // キー項目リストへのポインター
public:
  data_equal(const lpairkey *keylst0) : keylst(keylst0)
  // デフォールトコンストラクタ
  {}
  ~data_equal()                 // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_equal::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  bool equal = true;
  lpairkey::const_iterator kpos = keylst->begin();
  while (equal && kpos != keylst->end())
  {
    int idx = kpos->second - 1;
    if (kpos->first == dfdata)  // 数字データとして比較
    {
      double fdata1 = cdata1.fdata(idx);
      double fdata2 = cdata2.fdata(idx);
      if (fdata1 == BLANKf && fdata2 == BLANKf) // 数字に変換できない場合
      {
        if (cdata1.sdata(idx) != cdata2.sdata(idx)) equal = false;
      }
      else
      {
        if (sgndiff(fdata1, fdata2) != 0) equal = false;
      }
    }
    else                        // 文字データとして比較
    {
      if (cdata1.sdata(idx) != cdata2.sdata(idx)) equal = false;
    }
    if (equal) ++kpos;
  }
  return equal;                 // true:一致, false:不一致
}

class key_less                  // キー項目の比較用関数オブジェクト
{
private:
  const lpairkey *keylst;       // キー項目リストへのポインター
public:
  key_less(const lpairkey *keylst0) : keylst(keylst0)
  // コンストラクタ
  {}
  ~key_less()                   // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // キー項目を比較する
};

bool key_less::operator()(datac& cdata1, datac& cdata2) const
// キー項目を比較する
{
  int comp = 0;
  lpairkey::const_iterator kpos = keylst->begin();
  while (comp == 0 && kpos != keylst->end())
  {
    int idx = kpos->second - 1;
    if (kpos->first == dfdata)   // 数字データとして比較
    {
      double fdata1 = cdata1.fdata(idx);
      double fdata2 = cdata2.fdata(idx);
      if (fdata1 == BLANKf && fdata2 == BLANKf) // 数字に変換できない場合
      {
        comp = strcmp((cdata1.sdata(idx)).c_str(), (cdata2.sdata(idx)).c_str());
      }
      else
      {
        if (fdata1 < fdata2) comp = -1;
        else if (fdata1 > fdata2) comp = 1;
      }
    }
    else                        // 文字データとして比較
    {
      comp = strcmp((cdata1.sdata(idx)).c_str(), (cdata2.sdata(idx)).c_str());
    }
    if (comp == 0) ++kpos;
  }
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class sortdclass                // sortdメインクラス
{
private:
  string fname;                 // ファイル名
public:
  sortdclass() : fname()        // コンストラクタ
  {}
  ~sortdclass()                 // デストラクタ
  {}
  int get(const char* c_fname, const option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

int sortdclass::get(const char* c_fname, const option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void sortdclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) id.put(cout, formtag);
  if (opt.cd() == false) return;
  if (opt.keylst().size() > 0) datalst.sort(key_less(&(opt.keylst())));
  // ケースデータのソート
  if (opt.uniq() != uniqnon)
  {
    if (opt.uniq() == uniqlast) datalst.reverse(); // 最後の要素を残す場合
    datalst.unique(data_equal(&(opt.keylst())));
    if (opt.uniq() == uniqlast) datalst.reverse(); // 元に戻す
  }
  // キー項目が一致する重複ケースを削除する
  if (opt.sort() == false) datalst.reverse(); // 逆順ソート
  for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos) dpos->put(cout);
}

int main(int argc, char *argv[]) // メインルーチン
{
  sortdclass sortd;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) sortd.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { sortd.get(argv[i], opt, id, datalst);  fok = false; }
    }
    sortd.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
