// tser.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-11-05 16:58:36 tac>
//
// 縦並びの時系列データを横並びに変換する。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法:g++ -O -s -w tser.C -I/usr/local/include -L/usr/local/lib
//             -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tser  Version 7.41           Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:tser [<オプション> …] [<データファイル> …]",
  "機能:最初の項目をケースID項目、次の項目を時期項目、以後の項目を時系列デー",
  "     タ項目として、縦並びの時系列データを横並びにするんや。",
  "<オプション>",
  "  -cd[={-|+}]:ケースデータの出力指定なんや。(-cd:-cd=+)",
  "  -cid[={-|+}]:ケースID項目の欠損の取り使い指定なんや。(-cid:-cid=+)",
  "    -:欠損を除外  +:欠損を採用(オプション省略時)",
  "  -d=<時期間隔>:時系列データの時間間隔を指定するんや。(オプション省略時:-d=1)",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込むんや。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんや。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定なんや。(-id:-id=+)",
  "  -p=<時期数>:時系列データの時期数を指定するんや。",
  "    オプション省略時:-p=(開始時期-終了時期+1)",
  "  -t=<開始時期>-<終了時期>:時系列データの開始と終了時期を指定するんや。"
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vcd;                     // ケースデータ出力フラッグ
  bool vcid;                    // ケースID項目の欠損の取り使いフラッグ
  double vd;                    // 時期間隔
  bool vid;                     // データ識別情報出力フラッグ
  int vp;                       // 出力時期数
  int vpmax;                    // 最大時期数
  pairdbl vt;                   // 開始・終了時期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  bool cid() const       // ケースID項目の欠損の取り使いフラッグを返す
  { return vcid; }
  double d() const              // 時期間隔を返す
  { return vd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  int p() const                 // 出力時期数を返す
  { return vp; }
  int p(int p0)                 // 出力時期数を設定する
  { return (vp = p0); }
  int pmax() const              // 最大時期数を返す
  { return vpmax; }
  int pmax(int pmax0)           // 最大時期数を設定する
  { return (vpmax = pmax0); }
  double ts() const             // 開始時期を返す
  { return vt.first; }
  double ts(double ts0)         // 開始時期を設定する
  { return (vt.first = ts0); }
  double te() const             // 終了時期を返す
  { return vt.second; }
  double te(double te0)         // 終了時期を設定する
  { return (vt.second = te0); }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                 // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vcid(true), vd(1.0), vid(true), vpmax(0), vp(0), vt(BLANKf, BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cid")     // ケースID項目の欠損の取り使い指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcid = true;
    else if (opt[1] == "-") vcid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "d")       // 時期間隔の指定
  {
    vd = strtodf(opt[1]);
    if (vd <= 0.0 || vd == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "p")       // 時期数の指定
  {
    char* ptr = nullptr;
    vp = strtol((opt[1]).c_str(), &ptr, 0);
    if (vp < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "t")       // 開始時期と終了時期の指定
  {
    if (getrange(vt.first, vt.second, opt[1]) == 0) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check()            // オプションのチェックとセット
{
  if (vt.first == BLANKf || vt.second == BLANKf) return disperr(vpname, "-tオプションを指定してないやんか！");
  double pdouble = (vt.second - vt.first) / vd;
  if (pdouble > (double)INT_MAX) return disperr(vpname, "時期数が多すぎるやんか！");
  vpmax = (int)pdouble + 1;
  if (vpmax < vp) return disperr(vpname, "時期数が不適当やんか！");
  if (vp < 1) vp = vpmax;       // -pオプション省略時
  return 0;                     // エラーなし
}

class tserclass                 // tserメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id) const;
  // データ識別情報を出力する
  void stock_put(ostream& fout, const option& opt, const idd& id, datac& cdata, bool put) const;
  // ケースデータを時系列データとして保存・出力する
  void put_tdata(ostream& fout, const option& opt, vector<datac>& tcasevec) const;
  // 時系列データを出力する
public:
  tserclass() : fname()         // コンストラクタ
  {}
  ~tserclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void tserclass::put_id(ostream& fout, const option& opt, idd& id) const
// データ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  id.put_item(fout, 0, formtag); // ケース番号項目名出力
  id.put_item(fout, 1, formtag); // 時期項目名出力
  double ts = (opt.p() < opt.pmax()) ? 1.0 : opt.ts();
  double d = (opt.p() < opt.pmax()) ? 1.0 : opt.d();
  idd tid(id.title(), opt.p());
  for (size_t i = 2; i < id.ni(); ++i)
  {
    for (int j = 0; j < opt.p(); ++j)
    {
      ostringstream ostr;
      ostr << id.item(i) << " 時期" << ts + d * j;
      tid.assign_item(j, ostr.str());
    }
    tid.put_item(fout, formtag);
  }
}

void tserclass::stock_put(ostream& fout, const option& opt, const idd& id, datac& cdata, bool put) const
// ケースデータを時系列データとして保存・出力する
{
  double t = cdata.fdata(1);
  if (!put)                     // 無条件出力ではない場合
  {
    if ((!(opt.cid()) && cdata.sdata(0) == BLANKs) || t == BLANKf) return; // 欠損データ
    if (t < opt.ts() || opt.te() < t) return; // 時期項目範囲外データ
  }
  static vector<datac> tcasevec; // 時系列データ保存用ベクトル
  size_t tni = id.ni() - 2;
  if (tcasevec.empty())         // 初期設定
  {
    tcasevec.resize(tni);
    for (size_t i = 0; i < tni; ++i) (tcasevec[i]).resize(opt.pmax());
  }
  static string cid;            // ケースID保存用
  if (put || (!(cid.empty()) && cdata.sdata(0) != BLANKs && cid != cdata.sdata(0))) // 時系列データ出力
  {
    fout << cid << CHRdelm;
    put_tdata(fout, opt, tcasevec);
    cid.erase();
    for (size_t i = 0; i < tni; ++i)
    {
      for (int j = 0; j < opt.pmax(); ++j) (tcasevec[i]).assign(j, BLANKs);
    }
  }
  if (put) return;              // 無条件出力はデータ保存しない
  if (cid.empty()) cid = cdata.sdata(0);
  double tidx = (t - opt.ts()) / opt.d();
  int idx = (int)tidx;
  if (tidx != (double)idx) return; // 割り切れないときは保存しない
  for (size_t i = 0; i < tni; ++i) (tcasevec[i]).assign(idx, cdata.sdata(i + 2));
}

void tserclass::put_tdata(ostream& fout, const option& opt, vector<datac>& tcasevec) const
// 時系列データを出力する
{
  int jmin = 0;
  if (opt.p() < opt.pmax())     // 出力開始時期インデックスのサーチ
  {
    jmin = opt.pmax();
    for (size_t i = 0; i < tcasevec.size(); ++i)
    {
      int j = 0;
      while (j < opt.pmax() && (tcasevec[i]).sdata(j) == BLANKs) ++j;
      if (j < opt.pmax() && j < jmin) jmin = j;
    }
    if (jmin == opt.pmax()) jmin = 0;
  }
  fout << opt.ts() + opt.d() * jmin;
  for (size_t i = 0; i < tcasevec.size(); ++i)
  {
    int k = jmin;
    for (int j = 0; j < opt.p(); ++j, ++k) fout << CHRdelm << (tcasevec[i]).dstr(k);
  }
  fout << endl;
}

int tserclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (id.ni() < 3) throw runtime_error("時系列データ項目がないやんか！");
  if (opt.ts() == BLANKf || opt.te() == BLANKf) // 開始時期と終了時期の指定なし
  {
    ldata_itr pos = datalst.begin();
    double tmin = DBL_MAX, tmax = - DBL_MAX;
    for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos)
    {
      double t = dpos->fdata(1);
      if (t != BLANKf)
      {
        if (t < tmin) tmin = t;
        if (t > tmax) tmax = t;
      }
    }
    opt.ts(tmin);  opt.te(tmax);
  }
  if (opt.check() != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void tserclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return;
  ldata_itr pos = datalst.begin();
  while (pos != datalst.end())
  {
    stock_put(cout, opt, id, *pos, false);
    ++pos;
  }
  --pos;
  stock_put(cout, opt, id, *pos, true); // 最後に保存した時系列データを出力する
}

int main(int argc, char* argv[]) // メインルーチン
{
  tserclass tser;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) tser.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { tser.get(argv[i], opt, id, datalst);  fok = false; }
    }
    tser.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
