// tserc.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2024-07-21 08:08:44 tac>
//
// 横並びの時系列データを縦並びに変換する。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法:g++ -O -s -w tserc.C -I/usr/local/include -L/usr/local/lib
//             -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tserc  Version 7.41           Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:tserc [<オプション> …] [<データファイル> …]",
  "機能:最初の項目をケースID項目、以後の項目を時系列データ項目として、",
  "     横並びの時系列データを縦並びにするんや。",
  "<オプション>",
  "  -cd[={-|+}]:ケースデータの出力指定なんや。(-cd:-cd=+)",
  "  -d=<時期間隔>:時系列データの時間間隔を指定するんや。(オプション省略時:-d=1)",
  "  -f=<ファイル名>:指定したファイルからオプションを読み込むんや。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんや。(-h:-h=man)",
  "  -id[={-|+}]:データ識別情報の出力指定なんや。(-id:-id=+)",
  "  -nbg=<背景因子項目数>:背景因子項目数を指定するんや。(オプション省略時:-nbg=1)",
  "  -p=<時期数>:時系列データの時期数を指定するんや。(オプション省略時:全項目)",
  "  -ts=<開始時期>:時系列データの開始時期を指定するんや。(オプション省略時:-ts=0)"
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  bool vcd;                     // ケースデータ出力フラッグ
  double vd;                    // 時期間隔
  bool vid;                     // データ識別情報出力フラッグ
  int vnbg;                     // 背景因子項目数
  int vp;                       // 時期数
  double vts;                   // 開始時期
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  bool cd() const               // ケースデータ出力フラッグを返す
  { return vcd; }
  double d() const              // 時期間隔を返す
  { return vd; }
  bool id() const               // データ識別情報出力フラッグを返す
  { return vid; }
  int p() const                 // 出力時期数を返す
  { return vp; }
  int nbg() const               // 背景因子項目数を返す
  { return vnbg; }
  double ts() const             // 開始時期を返す
  { return vts; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);        // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcd(true), vd(1.0), vid(true), vnbg(1), vp(0), vts(0.0)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cd")           // ケースデータの出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vcd = true;
    else if (opt[1] == "-") vcd = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "d")       // 時期間隔の指定
  {
    vd = strtodf(opt[1]);
    if (vd <= 0.0 || vd == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "id")      // データ識別情報の出力指定
  {
    if (opt[1].empty() || opt[1] == "+" ) vid = true;
    else if (opt[1] == "-") vid = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "nbg")     // 背景因子項目数の指定
  {
    char* ptr = nullptr;
    vnbg = strtol((opt[1]).c_str(), &ptr, 0);
    if (vnbg < 0) err = badopt(vpname, optstr, USAGE, 0, INT_MAX);
  }
  else if (opt[0] == "p")       // 時期数の指定
  {
    char* ptr = nullptr;
    vp = strtol((opt[1]).c_str(), &ptr, 0);
    if (vp < 1) err = badopt(vpname, optstr, USAGE, 1, INT_MAX);
  }
  else if (opt[0] == "ts")      // 開始時期の指定
  {
    vts = strtodf(opt[1]);
    if (vts == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)   // オプションのチェックとセット
{
  if (vnbg + vp > (int)ni) return disperr(vpname, "項目数が不適当やんか！");
  if (vp < 1) vp = ni - vnbg;   // 時期数指定なしの時
  int nt = (ni - vnbg) / vp;    // 時期項目数
  if ((int)ni != vnbg + nt * vp) return disperr(vpname, "項目数が不適当やんか！");
  return 0;                     // エラーなし
}

class tsercclass                // tsercメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id) const;
  // データ識別情報を出力する
public:
  tsercclass() : fname()        // コンストラクタ
  {}
  ~tsercclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込みメモリに保存する
  void put(const option& opt, idd& id, ldata& datalst) const;
  // データを出力する
};

void tsercclass::put_id(ostream& fout, const option& opt, idd& id) const
// データ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  int nont = opt.nbg() + 1;      // 時系列項目以外の項目数
  int tni = nont + (id.ni() - opt.nbg()) / opt.p();
  idd tid(id.title(), tni);
  for (int bgidx = 0; bgidx < opt.nbg(); ++bgidx) tid.assign_item(bgidx, id.item(bgidx));
  tid.assign_item(opt.nbg(), "時期");
  for (int j = nont; j < tni; ++j)
  {
    int iidx = opt.nbg() + (j - nont) * opt.p();
    tid.assign_item(j, id.item(iidx));
  }
  tid.put_item(fout, formtag);
}

int tsercclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込みメモリに保存する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;
}

void tsercclass::put(const option& opt, idd& id, ldata& datalst) const
// データを出力する
{
  if (opt.id()) put_id(cout, opt, id);
  if (opt.cd() == false) return; // ケースデータ出力せず
  int nont = opt.nbg() + 1;      // 時系列項目以外の項目数
  int tni = nont + (id.ni() - opt.nbg()) / opt.p(); // 変換後の項目数
  datac tdata(tni);
  for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos)
  {
    for (int pi = 0; pi < opt.p(); ++pi)
    {
      for (int bgidx = 0; bgidx < opt.nbg(); ++bgidx) tdata.assign(bgidx, dpos->sdata(bgidx));
      tdata.assign(opt.nbg(), opt.ts() + pi * opt.d());
      for (int j = nont; j < tni; ++j)
      {
        int iidx = opt.nbg() + pi + (j - nont) * opt.p();
        tdata.assign(j, dpos->sdata(iidx));
      }
      tdata.put(cout);
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  tsercclass tserc;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) tserc.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { tserc.get(argv[i], opt, id, datalst);  fok = false; }
    }
    tserc.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
